from pathlib import Path
import pandas as pd

from fpdf.enums import Align, YPos
from ._design import PDFRounded as FPDF

from ._elements_analyzer import _add_multiplots, _add_outlier_analysis, _add_tableone

# Reports
def generate_analysis_report_pdf(
        outlier_analysis: str,
        multiplots: list,
        categorical_columns: list,
        continuous_columns: list,
        output_dir: str | Path
    ) -> None:
    """
    Generate a PDF report for the analysis, including plots, tables, and outlier analysis.

    Args:
        outlier_analysis (str): Text summary of outlier analysis to include in the report.
        multiplots (list): A list of paths to plots to include in the multiplots section.
        categorical_columns (list): A list of categorical columns to use for multiplots.
        continuous_columns (list): A list of continuous columns to use for multiplots.
        output_dir (str | Path): The directory where the generated PDF report will be saved.

    Returns:
        None: The function saves the generated PDF to the specified output directory.
    """
    output_dir = Path(output_dir)
    figures_dir = output_dir / 'figures'

    # Instantiate PDF
    pdf = FPDF()
    pdf.add_page()
    script_dir = Path(__file__).resolve().parent

    # Adding unicode fonts
    font_path = (script_dir / 'fonts/Inter_28pt-Regular.ttf')
    pdf.add_font("inter", style="", fname=font_path)
    font_path = (script_dir / 'fonts/Inter_28pt-Bold.ttf')
    pdf.add_font("inter", style="b", fname=font_path)
    
    # Title
    pdf.set_font('inter', 'B', 36)
    pdf.cell(text="jarvAIs Analyzer Report\n\n", new_y=YPos.NEXT) 

    # Add outlier analysis
    if outlier_analysis != '':
        pdf.set_y(pdf.get_y() + 10)
        pdf.set_font('inter', 'B', 24)
        pdf.cell(text="Outlier Analysis", new_y=YPos.NEXT)
        pdf = _add_outlier_analysis(pdf, outlier_analysis)

    # Add page-wide pairplots
    pdf.add_page()
    pdf.set_font('inter', 'B', 24)
    pdf.cell(h=pdf.t_margin, text='Pair Plot of Continuous Variables', new_y=YPos.NEXT)
    img = pdf.image((figures_dir / 'pairplot.png'), Align.C, h=pdf.eph*.5)
    pdf.set_y(pdf.t_margin + img.rendered_height + 25)

    # Add correlation plots
    pdf.set_font('inter', 'B', 24)
    pdf.cell(h=pdf.t_margin, text='Pearson and Spearman Correlation Plots', new_y=YPos.NEXT)

    corr_y = pdf.get_y() + 5

    pdf.image((figures_dir / 'pearson_correlation.png'), Align.L, corr_y, w=pdf.epw*.475)
    pdf.image((figures_dir / 'spearman_correlation.png'), Align.R, corr_y, w=pdf.epw*.475)

    # Add multiplots
    if multiplots and categorical_columns:
        pdf = _add_multiplots(pdf, multiplots, categorical_columns, continuous_columns)

    # Add demographic breakdown "table one"
    path_tableone = output_dir / 'tableone.csv'
    if path_tableone.exists():
        csv_df = pd.read_csv(path_tableone, na_filter=False).astype(str)
        pdf = _add_tableone(pdf, csv_df)

    # Save PDF
    pdf.output(output_dir / 'analysis_report.pdf')

def generate_explainer_report_pdf(
        problem_type: str,
        output_dir: str | Path
    ) -> None:
    """
    Generate a PDF report for the explainer with visualizations and metrics.

    This function creates a PDF report that includes plots and metrics 
    relevant to the specified problem type. The report is saved in the 
    specified output directory.

    Args:
        problem_type (str): The type of machine learning problem. 
            Supported values are 'binary', 'multiclass', 'regression', 
            and 'survival'.
        output_dir (str | Path): The directory where the generated PDF 
            report will be saved.

    Returns:
        None: The function saves the generated PDF to the specified output directory.
    """
    output_dir = Path(output_dir)
    figures_dir = output_dir / 'figures'

    # Instantiate PDF
    pdf = FPDF()
    pdf.add_page()
    script_dir = Path(__file__).resolve().parent

    # Adding unicode fonts
    font_path = (script_dir / 'fonts/Inter_28pt-Regular.ttf')
    pdf.add_font("inter", style="", fname=font_path)
    font_path = (script_dir / 'fonts/Inter_28pt-Bold.ttf')
    pdf.add_font("inter", style="b", fname=font_path)
    pdf.set_font('inter', '', 24)

    # Title
    pdf.write(5, "Explainer Report\n\n")

    pdf.image((figures_dir / 'test_metrics_bootstrap.png'), Align.C, h=pdf.eph//3.5, w=pdf.epw-20)
    pdf.image((figures_dir / 'validation_metrics_bootstrap.png'), Align.C, h=pdf.eph//3.5, w=pdf.epw-20)
    pdf.image((figures_dir /  'train_metrics_bootstrap.png'), Align.C, h=pdf.eph//3.5, w=pdf.epw-20)
    pdf.add_page()

    pdf.image((figures_dir / 'feature_importance.png'), Align.C, w=pdf.epw-20)
    pdf.add_page()

    if problem_type in ['binary', 'multiclass']:
        pdf.image((figures_dir / 'model_evaluation.png'), Align.C, w=pdf.epw-20)
        pdf.image((figures_dir / 'confusion_matrix.png'), Align.C, h=pdf.eph/2, w=pdf.epw-20)
        pdf.add_page()

        pdf.image((figures_dir / 'shap_barplot.png'), Align.C, h=pdf.eph/2, w=pdf.epw-20)
        pdf.image((output_dir /  'figures' / 'shap_heatmap.png'), Align.C, h=pdf.eph/2, w=pdf.epw-20)
    elif problem_type == 'regression':
        pdf.image((figures_dir / 'residual_plot.png'), Align.C, h=pdf.eph/2, w=pdf.epw-20)
        pdf.image((output_dir /  'figures' / 'true_vs_predicted.png'), Align.C, h=pdf.eph/2, w=pdf.epw-20)

    # Save PDF
    pdf.output((output_dir / 'explainer_report.pdf'))



"""
Helper functions and classes for aesthetic design.
"""

from typing import Callable
from fpdf import FPDF
from math import sqrt
from pathlib import Path
import seaborn as sns

import PIL
PIL.Image.MAX_IMAGE_PIXELS = None # Doing this to prevent PIL.Image.DecompressionBombError on large images

def config_plot(plot_type: str | None=None) -> Callable:
    """A decorator to configure plot settings using Seaborn and Matplotlib."""
    def decorator(func: Callable) -> Callable:
        def wrapper(*args, **kwargs):
            from matplotlib.font_manager import fontManager, FontProperties
            script_dir = Path(__file__).resolve().parent

            # Adding unicode fonts
            font_path = (script_dir / 'fonts/Inter_28pt-Regular.ttf')
            fontManager.addfont(font_path)
            inter = FontProperties(fname=font_path)
            rc_dict = {
                "figure.dpi": 300,
                'figure.subplot.left': 0.1,
                'figure.subplot.right': 0.9,
                'figure.subplot.bottom': 0.1,
                'figure.subplot.top': 0.9
            }
            if plot_type == 'multi':
                rc_dict['figure.subplot.left'] = 0.05
                rc_dict['figure.subplot.right'] = 0.95
                rc_dict['figure.subplot.bottom'] = 0.05
                rc_dict['figure.subplot.top'] = 0.95
                rc_dict['figure.subplot.hspace'] = .15
                rc_dict['figure.subplot.wspace'] = .15
            elif plot_type == 'ft':
                rc_dict['figure.subplot.left'] = 0.22
                rc_dict['figure.subplot.right'] = 0.98

            sns.set_theme(style="darkgrid", font=inter.get_name(), rc=rc_dict)
            
            return func(*args, **kwargs)
        return wrapper
    return decorator

class PDFRounded(FPDF):
    """A subclass of FPDF that adds support for drawing rounded rectangles."""
    def rounded_rect(
        self, x: float, y: float, w: float, h: float, r: float, 
        style: str = '', corners: str = '1234'
    ) -> None:
        """Draws a rounded rectangle."""
        k = self.k
        hp = self.h
        if(style=='F'):
            op='f'
        elif(style=='FD' or style=='DF'):
            op='B'
        else:
            op='S'
        myArc = 4/3 * (sqrt(2) - 1)
        self._out('%.2F %.2F m' % ((x+r)*k,(hp-y)*k))

        xc = x+w-r
        yc = y+r
        self._out('%.2F %.2F l' % (xc*k,(hp-y)*k))
        if '2' not in corners:
            self._out('%.2F %.2F l' % ((x+w)*k,(hp-y)*k))
        else:
            self._arc(xc + r*myArc, yc - r, xc + r, yc - r*myArc, xc + r, yc)

        xc = x+w-r
        yc = y+h-r
        self._out('%.2F %.2F l' % ((x+w)*k,(hp-yc)*k))
        if '3' not in corners:
            self._out('%.2F %.2F l' % ((x+w)*k,(hp-(y+h))*k))
        else:
            self._arc(xc + r, yc + r*myArc, xc + r*myArc, yc + r, xc, yc + r)

        xc = x+r
        yc = y+h-r
        self._out('%.2F %.2F l' % (xc*k,(hp-(y+h))*k))
        if '4' not in corners:
            self._out('%.2F %.2F l' % (x*k,(hp-(y+h))*k))
        else:
            self._arc(xc - r*myArc, yc + r, xc - r, yc + r*myArc, xc - r, yc)

        xc = x+r 
        yc = y+r
        self._out('%.2F %.2F l' % (x*k,(hp-yc)*k))
        if '1' not in corners:
            self._out('%.2F %.2F l' % (x*k,(hp-y)*k))
            self._out('%.2F %.2F l' % ((x+r)*k,(hp-y)*k))
        else:
            self._arc(xc - r, yc - r*myArc, xc - r*myArc, yc - r, xc, yc - r)
        self._out(op)
    
    def _arc(self, x1: float, y1: float, x2: float, y2: float, x3: float, y3: float) -> None:
        """Draws a Bézier curve arc for rounded corners."""    
        h = self.h
        self._out('%.2F %.2F %.2F %.2F %.2F %.2F c ' % (x1*self.k, (h-y1)*self.k,
            x2*self.k, (h-y2)*self.k, x3*self.k, (h-y3)*self.k))


import inspect, re
from typing import List
from pathlib import Path

from functools import partial
import pandas as pd
import numpy as np

import matplotlib.pyplot as plt
import seaborn as sns

from tabulate import tabulate
import fairlearn.metrics as fm

from sklearn.metrics import log_loss
import statsmodels.api as sm
from lifelines import CoxPHFitter

def infer_sensitive_features(data: pd.DataFrame) -> dict:
    """
    Infers potentially sensitive features from a DataFrame.
    """
    sensitive_keywords = [
        "gender", "sex", "age", "race", "ethnicity", "income",
        "religion", "disability", "nationality", "language",
        "marital", "citizen", "veteran", "status", "orientation",
        "disease", "regimen", "disease_site"
    ]

    sensitive_features = {
        col for col in data.columns 
        if any(re.search(rf"\b{keyword}\b", col, re.IGNORECASE) for keyword in sensitive_keywords)
    }

    return {sens_feat: data[sens_feat] for sens_feat in sensitive_features}

def get_metric(metric, sensitive_features=None):
    fn = getattr(fm, metric)
    params = inspect.signature(fn).parameters
    return partial(fn, sensitive_features=sensitive_features) if 'sensitive_features' in params and sensitive_features else fn

class BiasExplainer():
    """
    A class for explaining and analyzing bias in a predictive model's outcomes based on sensitive features.

    This class performs various fairness audits by evaluating predictive outcomes with respect to sensitive features such as
    gender, age, race, and more. It first runs statistical analyses using the OLS regression F-statistic p-value to assess any possibility 
    of bias in the model's predictions based on sensitive features. If the p-value is less than 0.05, indicating potential bias, 
    the class generates visualizations (such as violin plots) and calculates fairness metrics (e.g., demographic parity, equalized odds). 
    The results are presented for each sensitive feature, with optional relative fairness comparisons.

    Attributes:
        y_true (pd.DataFrame):
            The true target values for the model.
        y_pred (pd.DataFrame):
            The predicted values from the model.
        sensitive_features (dict or pd.DataFrame):
            A dictionary or DataFrame containing sensitive features used for fairness analysis.
        metrics (list):
            A list of metrics to calculate for fairness analysis. Defaults to ['mean_prediction', 'false_positive_rate', 'true_positive_rate'].
        mapper (dict):
            A dictionary mapping internal metric names to user-friendly descriptions.
        kwargs (dict):
            Additional parameters passed to various methods, such as metric calculation and plot generation.
    """
    def __init__(
            self, 
            y_true: pd.Series, 
            y_pred: np.ndarray, 
            sensitive_features: dict, 
            task: str,
            output_dir: Path,
            metrics: list = ['mean_prediction', 'false_positive_rate', 'true_positive_rate'], 
            **kwargs: dict
        ) -> None:
        self.y_true = y_true
        self.y_pred = y_pred
        self.task = task
        self.output_dir = output_dir
        self.mapper = {"mean_prediction": "Demographic Parity",
                       "false_positive_rate": "(FPR) Equalized Odds",
                       "true_positive_rate": "(TPR) Equalized Odds or Equal Opportunity"}
        self.metrics = metrics
        self.kwargs = kwargs
        
        # Convert sensitive_features to DataFrame or leave as Series
        if isinstance(sensitive_features, pd.DataFrame) or isinstance(sensitive_features, pd.Series):
            self.sensitive_features = sensitive_features
        elif isinstance(sensitive_features, dict):
            self.sensitive_features = pd.DataFrame.from_dict(sensitive_features)
        elif isinstance(sensitive_features, list):
            if any(isinstance(item, list) for item in sensitive_features):
                self.sensitive_features = pd.DataFrame(sensitive_features, columns=[f'sensitive_feature_{i}' for i in range(len(sensitive_features))])
            else:
                self.sensitive_features = pd.DataFrame(sensitive_features, columns=['sensitive_feature'])
        else:
            raise ValueError("sensitive_features must be a pandas DataFrame, Series, dictionary or list")
        
    def _generate_violin(self, sensitive_feature: str, bias_metric:np.ndarray) -> None:
        """Generate a violin plot for the bias metric."""
        plt.figure(figsize=(8, 6)) 
        sns.set_theme(style="whitegrid")  

        sns.violinplot(
            x=self.sensitive_features[sensitive_feature], 
            y=bias_metric, 
            palette="muted",  
            inner="quart", 
            linewidth=1.25 
        )

        bias_metric_name = 'log_loss' if self.task == 'binary' else 'root_mean_squared_error'

        plt.title(f'{bias_metric_name.title()} Distribution by {sensitive_feature}', fontsize=16, weight='bold')  
        plt.xlabel(f'{sensitive_feature}', fontsize=14)  
        plt.ylabel(f'{bias_metric_name.title()} per Patient', fontsize=14) 
        plt.xticks(rotation=45, ha='right')

        plt.tight_layout()  
        plt.savefig(self.output_dir / f'{sensitive_feature}_{bias_metric_name}.png') 
        plt.show()

    def _subgroup_analysis_OLS(self, sensitive_feature: str, bias_metric:np.ndarray) -> float:
        """Fit a statsmodels OLS model to the bias metric data, using the sensitive feature and print summary based on p_val."""
        one_hot_encoded = pd.get_dummies(self.sensitive_features[sensitive_feature], prefix=sensitive_feature)
        X_columns = one_hot_encoded.columns

        X = one_hot_encoded.values  
        y = bias_metric  

        X = sm.add_constant(X.astype(float), has_constant='add')
        model = sm.OLS(y, X).fit()

        if model.f_pvalue < 0.05:
            output = []

            print(f"⚠️  **Possible Bias Detected in {sensitive_feature.title()}** ⚠️\n")
            output.append(f"=== Subgroup Analysis for '{sensitive_feature.title()}' Using OLS Regression ===\n")

            output.append("Model Statistics:")
            output.append(f"    R-squared:                  {model.rsquared:.3f}")
            output.append(f"    F-statistic:                {model.fvalue:.3f}")
            output.append(f"    F-statistic p-value:        {model.f_pvalue:.4f}")
            output.append(f"    AIC:                        {model.aic:.2f}")
            output.append(f"    Log-Likelihood:             {model.llf:.2f}")

            summary_df = pd.DataFrame({
                'Feature': ['const'] + X_columns.tolist(),     # Predictor names (includes 'const' if added)
                'Coefficient': model.params,    # Coefficients
                'Standard Error': model.bse     # Standard Errors
            })
            table_output = tabulate(summary_df, headers='keys', tablefmt='grid', showindex=False, floatfmt=".3f")
            output.append("Model Coefficients:")
            output.append('\n'.join(['    ' + line for line in table_output.split('\n')]))

            output_text = '\n'.join(output)
            print(output_text)

            with open(self.output_dir / f'{sensitive_feature}_Cox_model_summary.txt', 'w') as f:
                f.write(output_text)

        return model.f_pvalue

    def _subgroup_analysis_CoxPH(self, sensitive_feature: str) -> None:
        """Fit a CoxPH model using the sensitive feature and print summary based on p_val."""
        one_hot_encoded = pd.get_dummies(self.sensitive_features[sensitive_feature], prefix=sensitive_feature)
        df_encoded = self.y_true.join(one_hot_encoded)

        cph = CoxPHFitter(penalizer=0.0001)
        cph.fit(df_encoded, duration_col='time', event_col='event')            
        
        if cph.log_likelihood_ratio_test().p_value < 0.05:
            output = []

            print(f"⚠️  **Possible Bias Detected in {sensitive_feature.title()}** ⚠️")
            output.append(f"=== Subgroup Analysis for '{sensitive_feature.title()}' Using Cox Proportional Hazards Model ===\n")

            output.append("Model Statistics:")
            output.append(f"    AIC (Partial):               {cph.AIC_partial_:.2f}")
            output.append(f"    Log-Likelihood:              {cph.log_likelihood_:.2f}")
            output.append(f"    Log-Likelihood Ratio p-value: {cph.log_likelihood_ratio_test().p_value:.4f}")
            output.append(f"    Concordance Index (C-index):   {cph.concordance_index_:.2f}")

            summary_df = pd.DataFrame({
                'Feature': cph.summary.index.to_list(),
                'Coefficient': cph.summary['coef'].to_list(),
                'Standard Error': cph.summary['se(coef)'].to_list()
            })
            table_output = tabulate(summary_df, headers='keys', tablefmt='grid', showindex=False, floatfmt=".3f")
            output.append("Model Coefficients:")
            output.append('\n'.join(['    ' + line for line in table_output.split('\n')]))

            output_text = '\n'.join(output)
            print(output_text)

            with open(self.output_dir / f'{sensitive_feature}_OLS_model_summary.txt', 'w') as f:
                f.write(output_text)

    def _calculate_fair_metrics(
            self, 
            sensitive_feature: str, 
            fairness_threshold: float, 
            relative: bool
        ) -> pd.DataFrame:
        """Calculate the Fairlearn metrics and return the results in a DataFrame."""
        _metrics = {metric: get_metric(metric, sensitive_features=self.sensitive_features[sensitive_feature]) for metric in self.metrics}
        metric_frame = fm.MetricFrame(
            metrics=_metrics, 
            y_true=self.y_true, 
            y_pred=self.y_pred, 
            sensitive_features=self.sensitive_features[sensitive_feature], 
            **self.kwargs
        )
        result = pd.DataFrame(metric_frame.by_group.T, index=_metrics.keys())
        result = result.rename(columns=self.mapper)

        if relative:
            largest_feature = self.sensitive_features[sensitive_feature].mode().iloc[0]
            results_relative = result.T / result[largest_feature]
            results_relative = results_relative.applymap(
                lambda x: f"{x:.3f} ✅" if x <= fairness_threshold or 1/x <= fairness_threshold 
                else f"{x:.3f} ❌")
            result = pd.concat([result, results_relative.T.rename(index=lambda x: f"Relative {x}")])
        
        return result
    
    def run(
            self, 
            relative: bool = False, 
            fairness_threshold: float = 1.2
        ) -> None:
        """
        Runs the bias explainer analysis on the provided data. It first evaluates the potential bias in the model's predictions
        using the OLS regression F-statistic p-value. If the p-value is below the threshold of 0.05, indicating 
        potential bias in the sensitive feature, the method proceeds to generate visualizations and calculate fairness metrics.

        Args:
            relative (bool): 
                If True, the metrics will be presented relative to the most frequent value of each sensitive feature.
            fairness_threshold (float): 
                A threshold for determining fairness based on relative metrics. If the relative metric exceeds this threshold, 
                a warning flag will be applied.
        """
        if self.task == 'binary':
            y_true_array = self.y_true.to_numpy()
            bias_metric = np.array([
                log_loss([y_true_array[idx]], [self.y_pred[idx]], labels=np.unique(y_true_array))
                for idx in range(len(y_true_array))
            ])
            self.y_pred = (self.y_pred >= .5).astype(int)
        elif self.task == 'regression':
            bias_metric = np.sqrt((self.y_true.to_numpy() - self.y_pred) ** 2)

        self.results = []
        for sensitive_feature in self.sensitive_features.columns:
            if self.task == 'survival':
                self._subgroup_analysis_CoxPH(sensitive_feature)
            else:
                f_pvalue = self._subgroup_analysis_OLS(sensitive_feature, bias_metric)
                if f_pvalue < 0.05:
                    self._generate_violin(sensitive_feature, bias_metric)
                    result = self._calculate_fair_metrics(sensitive_feature, fairness_threshold, relative)

                    print(f"\n=== Subgroup Analysis for '{sensitive_feature.title()}' using FairLearn ===\n")
                    table_output = tabulate(result.iloc[:, :4], headers='keys', tablefmt='grid')
                    print('\n'.join(['    ' + line for line in table_output.split('\n')]), '\n')

                    result.to_csv(self.output_dir / f'{sensitive_feature}_fm_metrics.csv')



from math import ceil, sqrt
from typing import Optional, Union

import numpy as np
import torch

TensorOrArray = Union[torch.Tensor, np.ndarray]

def normalize(data, mean=None, std=None, skip_cols=[]):
    """Normalizes the columns of Pandas DataFrame to zero mean and unit
    standard deviation."""
    if mean is None:
        mean = data.mean(axis=0)
    if std is None:
        std = data.std(axis=0)
    if skip_cols is not None:
        mean[skip_cols] = 0
        std[skip_cols] = 1
    return (data - mean) / std, mean, std

def encode_survival(time: Union[float, int, TensorOrArray],
                    event: Union[int, bool, TensorOrArray],
                    bins: TensorOrArray) -> torch.Tensor:
    """Encodes survival time and event indicator in the format
    required for MTLR training.

    For uncensored instances, one-hot encoding of binned survival time
    is generated. Censoring is handled differently, with all possible
    values for event time encoded as 1s. For example, if 5 time bins are used,
    an instance experiencing event in bin 3 is encoded as [0, 0, 0, 1, 0], and
    instance censored in bin 2 as [0, 0, 1, 1, 1]. Note that an additional
    'catch-all' bin is added, spanning the range `(bins.max(), inf)`.

    Parameters
    ----------
    time
        Time of event or censoring.
    event
        Event indicator (0 = censored).
    bins
        Bins used for time axis discretisation.

    Returns
    -------
    torch.Tensor
        Encoded survival times.
    """
    # TODO this should handle arrays and (CUDA) tensors
    if isinstance(time, (float, int, np.ndarray)):
        time = np.atleast_1d(time)
        time = torch.tensor(time)
    if isinstance(event, (int, bool, np.ndarray)):
        event = np.atleast_1d(event)
        event = torch.tensor(event)

    if isinstance(bins, np.ndarray):
        bins = torch.tensor(bins)

    try:
        device = bins.device
    except AttributeError:
        device = "cpu"

    time = np.clip(time, 0, bins.max())
    # add extra bin [max_time, inf) at the end
    y = torch.zeros((time.shape[0], bins.shape[0] + 1),
                    dtype=torch.float,
                    device=device)
    # For some reason, the `right` arg in torch.bucketize
    # works in the _opposite_ way as it does in numpy,
    # so we need to set it to True
    bin_idxs = torch.bucketize(time, bins, right=True)
    for i, (bin_idx, e) in enumerate(zip(bin_idxs, event)):
        if e == 1:
            y[i, bin_idx] = 1
        else:
            y[i, bin_idx:] = 1
    return y.squeeze()


def reset_parameters(model: torch.nn.Module) -> torch.nn.Module:
    """Resets the parameters of a PyTorch module and its children."""
    for m in model.modules():
        try:
            m.reset_parameters()
        except AttributeError:
            continue
    return model


def make_time_bins(times: TensorOrArray,
                   num_bins: Optional[int] = None,
                   use_quantiles: bool = True,
                   event: Optional[TensorOrArray] = None
                   ) -> torch.Tensor:
    """Creates the bins for survival time discretisation.

    By default, sqrt(num_observation) bins corresponding to the quantiles of
    the survival time distribution are used, as in https://github.com/haiderstats/MTLR.

    Parameters
    ----------
    times
        Array or tensor of survival times.
    num_bins
        The number of bins to use. If None (default), sqrt(num_observations)
        bins will be used.
    use_quantiles
        If True, the bin edges will correspond to quantiles of `times`
        (default). Otherwise, generates equally-spaced bins.
    event
        Array or tensor of event indicators. If specified, only samples where
        event == 1 will be used to determine the time bins.

    Returns
    -------
    torch.Tensor
        Tensor of bin edges.
    """
    # TODO this should handle arrays and (CUDA) tensors
    if event is not None:
        times = times[event == 1]
    if num_bins is None:
        num_bins = ceil(sqrt(len(times)))
    if use_quantiles:
        # NOTE we should switch to using torch.quantile once it becomes
        # available in the next version
        bins = np.unique(np.quantile(times, np.linspace(0, 1, num_bins)))
    else:
        bins = np.linspace(times.min(), times.max(), num_bins)
    bins = torch.tensor(bins, dtype=torch.float)
    return bins


def make_optimizer(opt_cls: torch.optim.Optimizer,
                   model: torch.nn.Module,
                   **kwargs) -> torch.optim.Optimizer:
    """Creates a PyTorch optimizer for MTLR training.

    This is a helper function to instantiate an optimizer with no weight decay
    on biases (which shouldn't be regularized) and MTLR parameters (which have
    a separate regularization mechanism). Note that the `opt_cls` argument
    should be the optimizer class, not an instantiated object (e.g. optim.Adam
    instead of optim.Adam(model.parameters(), ...)).

    Parameters
    ----------
    opt_cls
        The optimizer class to instantiate.
    model
        The PyTorch module whose parameters should be optimized.
    kwargs
        Additional keyword arguments to optimizer constructor.

    Returns
    -------
    torch.optim.Optimizer
        The instantiated optimizer object.

    """
    params_dict = dict(model.named_parameters())
    weights = [v for k, v in params_dict.items() if "mtlr" not in k and "bias" not in k]
    biases = [v for k, v in params_dict.items() if "bias" in k]
    mtlr_weights = [v for k, v in params_dict.items() if "mtlr_weight" in k]
    # Don't use weight decay on the biases and MTLR parameters, which have
    # their own separate L2 regularization
    optimizer = opt_cls([
        {"params": weights},
        {"params": biases, "weight_decay": 0.},
        {"params": mtlr_weights, "weight_decay": 0.},
    ], **kwargs)
    return optimizer



import pandas as pd
from mrmr import mrmr_classif, mrmr_regression
from sklearn.feature_selection import SelectKBest, VarianceThreshold, chi2, f_classif, f_regression

def var_reduction(X: pd.DataFrame, y: pd.DataFrame) -> pd.DataFrame:
    """Feature reduction with Variance Threshold."""
    selector = VarianceThreshold()
    _ = selector.fit_transform(X, y)
    selected_features = X.columns[selector.get_support(indices=True)]
    return X[selected_features]

def chi2_reduction(X: pd.DataFrame, y: pd.DataFrame, k: int) -> pd.DataFrame:
    """Feature reduction with K Best based on chi-squared statistic."""
    selector = SelectKBest(score_func=chi2, k=k)
    _ = selector.fit_transform(X, y)
    selected_features = X.columns[selector.get_support(indices=True)]
    return X[selected_features]

def kbest_reduction(task: str, X: pd.DataFrame, y: pd.DataFrame, k: int) -> pd.DataFrame:
    """Feature reduction with K Best based on F-statistic."""
    if task in ['binary', 'multiclass']:
        f_method = f_classif
    else:
        f_method = f_regression

    selector = SelectKBest(score_func=f_method, k=k)
    _ = selector.fit_transform(X, y)
    selected_features = X.columns[selector.get_support(indices=True)]
    return X[selected_features]

def mrmr_reduction(task: str, X: pd.DataFrame, y: pd.DataFrame, k: int) -> pd.DataFrame:
    """Feature reduction with mrmr."""
    mrmr_method = mrmr_classif if task in ['binary', 'multiclass'] else mrmr_regression
    selected_features = mrmr_method(X=X, y=y, K=k, n_jobs=1)
    return X[selected_features]



import logging
import warnings
from pathlib import Path

import pandas as pd
import yaml
from joblib import Parallel, delayed
from tableone import TableOne

from ._janitor import get_outliers, infer_types, replace_missing
from ..utils.pdf import generate_analysis_report_pdf
from ..utils.plot import (
    plot_corr,
    plot_frequency_table,
    plot_kaplan_meier_by_category,
    plot_one_multiplot,
    plot_pairplot,
    plot_umap,
)

logging.basicConfig(filename=(Path.cwd() / "warnings.log"), level=logging.INFO)

def custom_warning_handler(message, category, filename, lineno, file=None, line=None) -> None:
    log_message = f"{category.__name__}: {message} in {filename} at line {lineno}"
    logging.warning(log_message)

warnings.showwarning = custom_warning_handler

class Analyzer:
    """
    A data analysis and cleaning tool for preprocessing datasets, generating reports, and visualizations.

    Features:
        - Handles missing values and outliers.
        - Infers column types (categorical, continuous, date).
        - Supports one-hot encoding and survival analysis.
        - Generates summary statistics and correlation plots.
        - Produces a comprehensive PDF analysis report.

    Attributes:
        data (pd.DataFrame): Input dataset.
        target_variable (str, optional): Target variable in the dataset.
        task (str, optional): Type of analysis task.
        one_hot_encode (bool, optional): Whether to one-hot encode categorical columns.
        config (str | Path, optional): Path to a YAML configuration file.
        output_dir (str | Path, optional): Directory to save outputs. Default is the current directory.

    Example:
        ```python
        from jarvais.analyzer import Analyzer
        import pandas as pd

        data = pd.DataFrame({
            "age": [25, 32, 40],
            "income": [50000, 60000, 75000],
            "category": ["A", "B", "A"]
        })

        analyzer = Analyzer(data, target_variable="income", task="regression")
        analyzer.run()
        ```
    """
    def __init__(
            self,
            data: pd.DataFrame,
            target_variable: str | None = None,
            task: str | None = None,
            one_hot_encode: bool = False,
            config: str | Path = None,
            output_dir: str | Path = None
        ) -> None:

        self.data = data
        self.target_variable = target_variable
        self.task = task
        self.one_hot_encode = one_hot_encode

        assert_message = "When setting task to 'survival', target_variable must be 'event' and 'time' must be in data"
        if self.task == 'survival':
            assert target_variable == 'event' and 'time' in data.columns, assert_message

        self.output_dir = Path.cwd() if output_dir is None else Path(output_dir)
        self.output_dir.mkdir(parents=True, exist_ok=True)

        if config is not None:
            config = Path(config)
            if config.is_file():
                with config.open('r') as file:
                    self.config = yaml.safe_load(file)
            else:
                raise ValueError(f'Config file does not exist at {config}')
        else:
            self.config = config

        self.outlier_analysis = '' # Used later when writing to PDF

    def _create_config(self) -> None:
        """
        Create and save a configuration file for column types, outlier handling, and missing value strategies.

        Steps:
        1. **Infer Column Types**: Identifies categorical, continuous, and date columns using `infer_types`.
        2. **Handle NaN Columns**: Drops columns entirely filled with NaN and updates the continuous column list.
        3. **Outlier Detection**: Identifies outliers in categorical columns and stores the mappings.
        4. **Missing Value Strategy**: Sets default imputation strategies for categorical and continuous variables.
        """
        print('Config file not found. Creating custom...')

        self.config = {}
        columns = {}

        self.categorical_columns, self.continuous_columns, self.date_columns = infer_types(self.data)
        # Replace all non numerical values with NaN
        self.data[self.continuous_columns] = self.data[self.continuous_columns].apply(pd.to_numeric, errors='coerce')

        nan_ = self.data.apply(lambda col: col.isna().all())
        nan_columns = nan_[nan_].index.tolist()
        if len(nan_columns) > 0:
            print("Columns that are all NaN(probably ID columns) dropping...: ", nan_columns)
            self.continuous_columns = list(set(self.continuous_columns) - set(nan_columns))

        print("Used a heuristic to define categorical and continuous columns. Please review!")
        
        columns['categorical'] = self.categorical_columns
        columns['continuous'] = self.continuous_columns
        columns['date'] = self.date_columns
        columns['other'] = nan_columns

        self.config['columns'] = columns

        outlier_analysis, mapping = get_outliers(self.data, self.categorical_columns)

        self.outlier_analysis += outlier_analysis
        self.config['mapping'] = mapping

        self.config['missingness_strategy'] = {}
        # Defining default replacement for each missing categorical variable
        self.config['missingness_strategy']['categorical'] = {cat :'Unknown' for cat in self.categorical_columns}
        # Defining default replacement for each missing continuous variable
        self.config['missingness_strategy']['continuous'] = {cont :'median' for cont in self.continuous_columns}

    def _apply_config(self) -> None:

        print('Applying changes from config...\n')

        for key in self.config['mapping'].keys():
            assert key in self.data.columns, f"{key} in mapping file not found data"
            self.data.loc[:, [key]] = self.data.loc[:, key].replace(self.config['mapping'][key])

        self.data = replace_missing(self.data, self.categorical_columns, self.continuous_columns, self.config)

    def _create_multiplots(self, figures_dir: Path) -> None:
        """Generate and save multiplots for each categorical variable against all continuous variables."""
        self.multiplots = [] # Used to save in PDF later

        (figures_dir / 'multiplots').mkdir(parents=True, exist_ok=True)

        self.multiplots = Parallel(n_jobs=-1)(
            delayed(plot_one_multiplot)(
                self.data,
                self.umap_data,
                var,
                self.continuous_columns,
                figures_dir
            ) for var in self.categorical_columns
        )

    def run(self) -> None:
        """Run the data cleaning and visualization process."""
        if self.config is None:
            self._create_config()
        else:
            self.continuous_columns = self.config['columns']['continuous']
            self.categorical_columns = self.config['columns']['categorical']
            # Replace all non numerical values with NaN
            self.data[self.continuous_columns] = self.data[self.continuous_columns].apply(pd.to_numeric, errors='coerce')
            self.outlier_analysis, _ = get_outliers(self.data, self.categorical_columns)

        print(f"Feature Types:\n  - Categorical: {self.categorical_columns}\n  - Continuous: {self.continuous_columns}")
        print(f"\n\nOutlier Analysis:\n{self.outlier_analysis}")

        with open(self.output_dir / 'config.yaml', 'w') as f:
            yaml.dump(self.config, f)

        self._apply_config()

        # Create Table One
        df_keep = self.data[self.continuous_columns + self.categorical_columns]

        self.mytable = TableOne(df_keep, categorical=self.categorical_columns, pval=False)
        print(self.mytable.tabulate(tablefmt = "grid"))
        self.mytable.to_csv(self.output_dir / 'tableone.csv')

        # PLOTS
        figures_dir = self.output_dir / 'figures'
        figures_dir.mkdir(exist_ok=True, parents=True)

        if self.task == 'survival':
            data_x = self.data.drop(columns=['time', 'event'])
            data_y = self.data[['time', 'event']]
            categorical_columns  = [cat for cat in self.categorical_columns if cat != 'event']
            plot_kaplan_meier_by_category(
                data_x, data_y,
                categorical_columns,
                figures_dir / 'kaplan_meier'
            )

        # Correlation Plots
        p_corr = self.data[self.continuous_columns].corr(method="pearson")
        s_corr = self.data[self.continuous_columns].corr(method="spearman")
        size = 1 + len(self.continuous_columns)*1.2
        plot_corr(p_corr, size, file_name='pearson_correlation.png', output_dir=figures_dir, title="Pearson Correlation")
        plot_corr(s_corr, size, file_name='spearman_correlation.png', output_dir=figures_dir, title="Spearman Correlation")

        # Categorical cross frequency table
        plot_frequency_table(self.data, self.categorical_columns, figures_dir)

        # UMAP reduced data + Plots
        self.umap_data = plot_umap(self.data, self.continuous_columns, figures_dir)

        # Plot pairplot: keeping only the top ten correlated pairs in the pair plot
        if self.target_variable in self.categorical_columns:
            plot_pairplot(self.data, self.continuous_columns, output_dir=figures_dir, target_variable=self.target_variable)
        else:
            plot_pairplot(self.data, self.continuous_columns, output_dir=figures_dir)

        # Create Multiplots
        self._create_multiplots(figures_dir)

        if self.one_hot_encode:
            self.data = pd.get_dummies(
                self.data,
                columns=[cat for cat in self.categorical_columns if cat != self.target_variable],
                dtype=float,
                prefix_sep='|' # Using this to make it obvious OHE features
            )

        self.data.to_csv(self.output_dir / 'updated_data.csv')

        # Create Output PDF
        generate_analysis_report_pdf(
            self.outlier_analysis,
            self.multiplots,
            self.categorical_columns,
            self.continuous_columns,
            self.output_dir
        )

    @classmethod
    def dry_run(cls, data: pd.DataFrame) -> dict:
        """Simply returns generated config and displays TableOne."""
        analyzer = cls(data)
        analyzer._create_config()

        df_keep = analyzer.data[analyzer.continuous_columns + analyzer.categorical_columns]

        print(f"\n\nFeature Types:\n  - Categorical: {analyzer.categorical_columns}\n  - Continuous: {analyzer.continuous_columns}")
        print(f"\n\nOutlier Analysis:\n{analyzer.outlier_analysis}")

        mytable = TableOne(df_keep, categorical=analyzer.categorical_columns, pval=False)
        print(mytable.tabulate(tablefmt = "grid"))

        return analyzer.config





import yaml
import pickle
from pathlib import Path
from typing import List

import numpy as np
import pandas as pd
from autogluon.core.metrics import make_scorer
from autogluon.tabular import TabularPredictor
from autogluon.tabular.configs.hyperparameter_configs import get_hyperparameter_config
from sklearn.model_selection import train_test_split
from tabulate import tabulate

from ._feature_reduction import chi2_reduction, kbest_reduction, mrmr_reduction, var_reduction
from ._leaderboard import format_leaderboard
from ._simple_regression_model import SimpleRegressionModel
from ._training import train_autogluon_with_cv, train_survival_models
from ..explainer import Explainer
from ..utils.functional import auprc
from .survival import LitDeepSurv, LitMTLR

class TrainerSupervised:
    """
    TrainerSupervised class for supervised jarvAIs workflows.

    This class provides functionality for feature reduction, training models (e.g., AutoGluon, survival models), 
    and performing inference. It supports various tasks such as binary/multiclass classification, regression, 
    and survival analysis.

    Attributes:
        task (str, optional): Type of task. Must be one of {'binary', 'multiclass', 'regression', 'survival'}. 
        reduction_method (str | None, optional): Feature reduction method. Supported methods include 
            {'mrmr', 'variance_threshold', 'corr', 'chi2'}.
        keep_k (int, optional): Number of features to retain during reduction.
        output_dir (str | Path, optional): Directory for saving outputs. Defaults to the current working directory.
        
    Example:
        ```python
        from jarvais.trainer import TrainerSupervised

        trainer = TrainerSupervised(
            task="binary",
            reduction_method="mrmr",
            keep_k=10,
            output_dir="./results"
        )
        trainer.run(data=my_data, target_variable="target")
        
        predictions = trainer.infer(new_data)
        ```
    """
    def __init__(
            self,
            task: str=None,
            reduction_method: str | None = None,
            keep_k: int = 2,
            output_dir: str | Path = None
        ) -> None:

        self.task = task
        self.reduction_method = reduction_method
        self.keep_k = keep_k

        if task not in ['binary', 'multiclass', 'regression', 'survival', None]:
            raise ValueError("Invalid task parameter. Choose one of: 'binary', 'multiclass', 'regression', 'survival'. Providing None defaults to Autogluon infering.")

        self.output_dir = Path.cwd() if output_dir is None else Path(output_dir)
        self.output_dir.mkdir(exist_ok=True, parents=True)

    def _feature_reduction(self, X: pd.DataFrame, y: pd.DataFrame | pd.Series) -> pd.DataFrame:
        """
        Reduce features based on the specified reduction method. 
        
        One-hot encoding applied before reduction and reverted afterward.
        """
        # Step 1: Identify categorical columns
        categorical_columns = X.select_dtypes(include=['object', 'category']).columns.tolist()

        mappin = {}

        def find_category_mappings(df, variable):
            return {k: i for i, k in enumerate(df[variable].dropna().unique(), 0)}

        def integer_encode(df, variable, ordinal_mapping):
            df[variable] = df[variable].map(ordinal_mapping)

        for variable in categorical_columns:
            mappings = find_category_mappings(X, variable)
            mappin[variable] = mappings

        for variable in categorical_columns:
            integer_encode(X, variable, mappin[variable])

        # Step 3: Perform feature reduction
        if self.reduction_method == 'mrmr':
            X_reduced = mrmr_reduction(self.task, X, y, self.keep_k)
        elif self.reduction_method == 'variance_threshold':
            X_reduced = var_reduction(X, y)
        elif self.reduction_method == 'corr':
            X_reduced = kbest_reduction(self.task, X, y, self.keep_k)
        elif self.reduction_method == 'chi2':
            if self.task not in ['binary', 'multiclass']:
                raise ValueError('chi-squared reduction can only be done with classification tasks')
            X_reduced = chi2_reduction(X, y, self.keep_k)
        else:
            raise ValueError('Unsupported reduction method: {}'.format(self.reduction_method))

        for col in categorical_columns:
            if col in X_reduced.columns:
                inv_map = {v: k for k, v in mappin[col].items()}
                X_reduced[col] = X_reduced[col].map(inv_map)

        return X_reduced

    def _train_autogluon_with_cv(self) -> None:
        self.predictors, leaderboard, self.best_fold, self.X_val, self.y_val = train_autogluon_with_cv(
            pd.concat([self.X_train, self.y_train], axis=1),
            pd.concat([self.X_test, self.y_test], axis=1),
            target_variable=self.target_variable,
            task=self.task,
            extra_metrics=self.extra_metrics,
            eval_metric=self.eval_metric,
            num_folds=self.k_folds,
            output_dir=(self.output_dir / 'autogluon_models'),
            **self.kwargs
        )

        self.predictor = self.predictors[self.best_fold]
        self.trainer_config['best_fold'] = self.best_fold

        # Update train data to remove validation
        self.X_train = self.X_train[~self.X_train.index.isin(self.X_val.index)]
        self.y_train = self.y_train[~self.y_train.index.isin(self.y_val.index)]

        print('\nModel Leaderboard (Displays values in "mean [min, max]" format across training folds)\n------------------------------------------------------------------------------------')
        print(tabulate(
            leaderboard.sort_values(by='score_test', ascending=False)[self.show_leaderboard],
            tablefmt = "grid",
            headers="keys",
            showindex=False
        ))

    def _train_autogluon(self) -> None:
        self.predictor = TabularPredictor(
            label=self.target_variable, 
            problem_type=self.task, 
            eval_metric=self.eval_metric,
            path=(self.output_dir / 'autogluon_models' / 'autogluon_models_best_fold'),
            log_to_file=False,
        ).fit(
            pd.concat([self.X_train, self.y_train], axis=1),
            **self.kwargs
        )

        self.X_val, self.y_val = self.predictor.load_data_internal(data='val', return_y=True)
        # Update train data to remove validation
        self.X_train = self.X_train[~self.X_train.index.isin(self.X_val.index)]
        self.y_train = self.y_train[~self.y_train.index.isin(self.y_val.index)]

        train_leaderboard = self.predictor.leaderboard(
            pd.concat([self.X_train, self.y_train], axis=1),
            extra_metrics=self.extra_metrics).round(2)
        val_leaderboard = self.predictor.leaderboard(
            pd.concat([self.X_val, self.y_val], axis=1),
            extra_metrics=self.extra_metrics).round(2)
        test_leaderboard = self.predictor.leaderboard(
            pd.concat([self.X_test, self.y_test], axis=1),
            extra_metrics=self.extra_metrics).round(2)

        leaderboard = pd.merge(
            pd.merge(
                format_leaderboard(train_leaderboard, self.extra_metrics, 'score_train'),
                format_leaderboard(val_leaderboard, self.extra_metrics, 'score_val'),
                on='model'
            ),
            format_leaderboard(test_leaderboard, self.extra_metrics, 'score_test'),
            on='model'
        )

        print('\nModel Leaderboard\n----------------')
        print(tabulate(
            leaderboard.sort_values(by='score_test', ascending=False)[self.show_leaderboard],
            tablefmt = "grid",
            headers="keys",
            showindex=False))

    def run(
            self,
            data: pd.DataFrame,
            target_variable: str,
            test_size: float = 0.2,
            exclude: List[str] | None = None,
            stratify_on: str | None = None,
            explain: bool = False,
            k_folds: int = 5,
            **kwargs:dict
        ) -> None:
        """
        Execute the jarvAIs Trainer pipeline on the given dataset.

        Args:
            data (pd.DataFrame): The input dataset containing features and target.
            target_variable (str): The name of the target variable in the dataset.
            test_size (float, optional): Proportion of the dataset to include in the test split. 
                Must be between 0 and 1. Default is 0.2.
            exclude (list of str, optional): List of columns to exclude from the feature set. 
                Default is an empty list.
            stratify_on (str, optional): Column to use for stratification, if any. 
                Must be compatible with `target_variable`.
            explain (bool, optional): Whether to generate explainability reports for the model. 
                Default is False.
            k_folds (int, optional): Number of folds for cross-validation. If 1, uses AutoGluon-specific validation. 
                Default is 5.
            kwargs (dict, optional): Additional arguments passed to the AutoGluon predictor's `fit` method.
        """
        self.trainer_config = dict()
        self.trainer_config['task'] = self.task
        self.trainer_config['output_dir'] = self.output_dir.as_posix()

        self.target_variable = target_variable
        self.trainer_config['target_variable'] = target_variable
        self.k_folds = k_folds
        self.trainer_config['k_folds'] = k_folds
        self.kwargs = kwargs

        self.trainer_config['test_size'] = test_size
        self.trainer_config['stratify_on'] = stratify_on

        # Initialize mutable defaults
        if exclude is None:
            exclude = []

        if isinstance(target_variable, list): # Happens for survival data
            exclude += target_variable
        else:
            exclude.append(target_variable)

        try:
            X = data.drop(columns=exclude)
            y = data[target_variable]
        except KeyError as e:
            raise ValueError(f"Invalid column specified: {e}")

        # Optional feature reduction
        if getattr(self, "reduction_method", None):
            print(f"Applying {self.reduction_method} for feature reduction")
            X = self._feature_reduction(X, y)
            print(f"Features retained: {list(X.columns)}")

            self.feature_names = list(X.columns)
            self.trainer_config['reduction_method'] = self.reduction_method
            self.trainer_config['reduced_feature_set'] = self.feature_names

        if self.task in {'binary', 'multiclass'}:
            stratify_col = (
                y.astype(str) + '_' + data[stratify_on].astype(str)
                if stratify_on is not None
                else y
            )
        else:
            stratify_col = None

        self.X_train, self.X_test, self.y_train, self.y_test = train_test_split(
            X, y, test_size=test_size, stratify=stratify_col, random_state=42)

        if self.task == 'survival':
            self.predictors, scores, data_train, data_val = train_survival_models(
                self.X_train, 
                self.y_train, 
                self.X_test, 
                self.y_test, 
                self.output_dir
            )
            self.predictor = self.predictors[max(scores, key=scores.get)]
            self.trainer_config['survival_models_info'] = scores

            self.X_train, self.y_train = data_train.drop(columns=['time', 'event']), data_train[['time', 'event']] 
            self.X_val, self.y_val = data_val.drop(columns=['time', 'event']), data_val[['time', 'event']] 
        else:
            (self.output_dir / 'autogluon_models').mkdir(exist_ok=True, parents=True)

            if self.task in ['binary', 'multiclass']:
                self.eval_metric = 'roc_auc'
            elif self.task == 'regression':
                self.eval_metric = 'r2'

            ag_auprc_scorer = make_scorer(
                name='auprc', # Move this to a seperate file?
                score_func=auprc,
                optimum=1,
                greater_is_better=True,
                needs_class=True)

            # When changing extra_metrics consider where it's used and make updates accordingly
            self.extra_metrics = ['f1', ag_auprc_scorer] if self.task in ['binary', 'multiclass'] else ['root_mean_squared_error']
            self.show_leaderboard = ['model', 'score_test', 'score_val', 'score_train']

            custom_hyperparameters = get_hyperparameter_config('default')
            custom_hyperparameters[SimpleRegressionModel] = {}
            kwargs['hyperparameters'] = custom_hyperparameters

            if k_folds > 1:
                self._train_autogluon_with_cv()
            else:
                self._train_autogluon()
        
        self.trained = True

        self.data_dir = self.output_dir / 'data'
        self.data_dir.mkdir(parents=True, exist_ok=True)
        self.X_train.to_csv((self.data_dir / 'X_train.csv'), index=False)
        self.X_test.to_csv((self.data_dir / 'X_test.csv'), index=False)
        self.X_val.to_csv((self.data_dir / 'X_val.csv'), index=False)
        self.y_train.to_csv((self.data_dir / 'y_train.csv'), index=False)
        self.y_test.to_csv((self.data_dir / 'y_test.csv'), index=False)
        self.y_val.to_csv((self.data_dir / 'y_val.csv'), index=False)

        with (self.output_dir / 'trainer_config.yaml').open('w') as f:
            yaml.dump(self.trainer_config, f)

        if explain:
            explainer = Explainer.from_trainer(self)
            explainer.run()

    def model_names(self) -> List[str]:
        """
        Returns all trainer model names.

        This method retrieves the names of all models associated with the 
        current predictor. It requires that the predictor has been trained.

        Returns:
            list: A list of model names available in the predictor.

        Raises:
            ValueError: If the model has not been trained (`self.trained` is False).
        """
        if not self.trained:
            raise ValueError("The model must be trained before accessing model names.")

        if self.task == 'survival':
            return list(self.predictors.keys())
        else:        
            return self.predictor.model_names()

    def infer(self, data: pd.DataFrame, model: str = None) -> np.ndarray:
        """
        Perform inference using the trained predictor on the provided data.

        This method generates predictions based on the input data using the 
        specified model. If no model is provided, the default model is used. 
        The predictor must be trained before inference can be performed.

        Args:
            data (pd.DataFrame): The input data for which inference is to be performed.
            model (str, optional): The name of the model to use for inference. 
                If None, the default model is used.

        Returns:
            np.ndarray: The prediction results from the model.

        Raises:
            ValueError: If the model has not been trained (`self.trained` is False).
            ValueError: If the specified model name is not found in the predictor.
        """
        if not self.trained:
            raise ValueError("The model must be trained before performing inference.")
        if not model is None and not model in self.model_names():
            raise ValueError(f"Model '{model}' not in trainer. Use model_names() to list valid available models.")

        if self.task == 'survival':
            if model is None:
                inference = self.predictor.predict(data)
            else:
                inference = self.predictors[model].predict(data)
        else:
            if self.predictor.can_predict_proba:
                inference = self.predictor.predict_proba(data, model, as_pandas=False)[:, 1]
            else:
                inference = self.predictor.predict(data, model, as_pandas=False)

        return inference
    
    @classmethod
    def load_trainer(cls, project_dir: str | Path):
        """
        Load a trained TrainerSupervised from the specified directory.

        Args:
            project_dir (str or Path, optional): The directory where the trainer was run.

        Returns:
            trainer (TrainerSupervised): The loaded Trainer.
        """
        project_dir = Path(project_dir)
        with (project_dir / 'trainer_config.yaml').open('r') as f:
            trainer_config = yaml.safe_load(f)

        trainer = cls()
        trainer.task = trainer_config['task']
        trainer.output_dir = project_dir
        
        if trainer.task == 'survival':
            model_dir = (project_dir / 'survival_models')
            
            trainer.predictors = {}
            model_info = trainer_config['survival_models_info']
            for model_name, _ in model_info.items():
                if model_name == 'MTLR':
                    trainer.predictors[model_name] = LitMTLR.load_from_checkpoint(model_dir / "MTLR.ckpt")
                elif model_name == 'DeepSurv':
                    trainer.predictors[model_name] = LitDeepSurv.load_from_checkpoint(model_dir / "DeepSurv.ckpt")
                else:
                    with (model_dir / f'{model_name}.pkl').open("rb") as f:
                        trainer.predictors[model_name] = pickle.load(f)

            trainer.predictor = trainer.predictors[max(model_info, key=model_info.get)]
        else:
            model_dir = (project_dir / 'autogluon_models' / 'autogluon_models_best_fold')
            trainer.predictor = TabularPredictor.load(model_dir, verbosity=1)

        trainer.trained = True
        
        trainer.X_test = pd.read_csv(project_dir / 'data' / 'X_test.csv')
        trainer.X_val = pd.read_csv(project_dir / 'data' / 'X_val.csv')
        trainer.X_train = pd.read_csv(project_dir / 'data' / 'X_train.csv')
        trainer.y_test = pd.read_csv(project_dir / 'data' / 'y_test.csv').squeeze()
        trainer.y_val = pd.read_csv(project_dir / 'data' / 'y_val.csv').squeeze()
        trainer.y_train = pd.read_csv(project_dir / 'data' / 'y_train.csv').squeeze()
  
        return trainer





import pandas as pd
from numpy import float32, ndarray

from autogluon.core.models import AbstractModel
from autogluon.features.generators import LabelEncoderFeatureGenerator

from sklearn.linear_model import LinearRegression, LogisticRegression

class SimpleRegressionModel(AbstractModel):
    """
    A simple regression model wrapped in an AutoGluon model class.

    It can handle `binary`, `multiclass`, `regression` tasks.
    """
    def __init__(self, **kwargs):
        # Simply pass along kwargs to parent, and init our internal `_feature_generator` variable to None
        super().__init__(**kwargs)
        self._feature_generator = None

    # The `_preprocess` method takes the input data and transforms it to the internal representation usable by the model.
    # `_preprocess` is called by `preprocess` and is used during model fit and model inference.
    def _preprocess(self, X: pd.DataFrame, is_train=False, **kwargs) -> ndarray:
        # print(f'Entering the `_preprocess` method: {len(X)} rows of data (is_train={is_train})')
        X = super()._preprocess(X, **kwargs)

        if is_train:
            # X will be the training data.
            self._feature_generator = LabelEncoderFeatureGenerator(verbosity=0)
            self._feature_generator.fit(X=X)
        if self._feature_generator.features_in:
            # This converts categorical features to numeric via stateful label encoding.
            X = X.copy()
            X[self._feature_generator.features_in] = self._feature_generator.transform(X=X)

        # Add a fillna call to handle missing values.
        return X.fillna(0).to_numpy(dtype=float32)

    # The `_fit` method takes the input training data (and optionally the validation data) and trains the model.
    def _fit(self,
            X: pd.DataFrame,  # training data
            y: pd.Series,  # training labels
            **kwargs):

        # Store the feature names before transforming to numpy
        feature_names = X.columns if isinstance(X, pd.DataFrame) else range(X.shape[1])

        # Make sure to call preprocess on X near the start of `_fit`.
        X = self.preprocess(X, is_train=True)

        # This fetches the user-specified (and default) hyperparameters for the model.
        params = self._get_model_params()

        if self.problem_type == 'regression':
            self.model = LinearRegression()
        else:
            self.model = LogisticRegression(**params)
        self.model.fit(X, y)

        # Print the coefficients and the intercept after training
        coefs = self.model.coef_.flatten()
        intercept = self.model.intercept_ if self.problem_type == 'regression' else self.model.intercept_[0]

        # Create a DataFrame to display feature names with their corresponding coefficients
        coef_df = pd.DataFrame({
            'Feature': feature_names,
            'Coefficient': coefs
        })

        # Print the coefficients along with their corresponding features
        # print("\nSimple Regression Model Coefficients:")
        # print(coef_df.sort_values(by='Coefficient', key=abs, ascending=False).to_string(index=False))  # sort by coefficient value for better readability

        # # Print the intercept separately
        # print(f"\nSimple Regression Model Intercept: {intercept}")

        # print('Exiting the `_fit` method')

    # The `_set_default_params` method defines the default hyperparameters of the model.
    def _set_default_params(self):
        default_params = {
            'solver': 'lbfgs',  # Solver to use in the optimization problem
            'max_iter': 1000,    # Maximum number of iterations for convergence
            'random_state': 0,   # Set the random seed
        }
        for param, val in default_params.items():
            self._set_default_param_value(param, val)

    # The `_get_default_auxiliary_params` method defines model-agnostic parameters such as maximum memory usage and valid input column dtypes.
    def _get_default_auxiliary_params(self) -> dict:
        default_auxiliary_params = super()._get_default_auxiliary_params()
        extra_auxiliary_params = dict(
            valid_raw_types=['int', 'float', 'category'],
        )
        default_auxiliary_params.update(extra_auxiliary_params)
        return default_auxiliary_params



import pytest
import shutil
import pandas as pd
from pathlib import Path
from sklearn.datasets import make_classification, make_regression
from jarvais.explainer import Explainer
from jarvais.trainer import TrainerSupervised

@pytest.fixture
def classification_data():
    X, y = make_classification(
        n_samples=100, n_features=5, n_informative=3, n_redundant=1, random_state=42,
    )
    X = pd.DataFrame(X, columns=[f"feature{i}" for i in range(1, X.shape[1] + 1)])
    y = pd.Series(y, name="target")
    return X, y

@pytest.fixture
def regression_data():
    X, y = make_regression(
        n_samples=50, n_features=5, noise=0.1, random_state=42
    )
    X = pd.DataFrame(X, columns=[f"feature{i}" for i in range(1, X.shape[1] + 1)])
    y = pd.Series(y, name="target")
    return X, y

@pytest.fixture
def tmpdir():
    temp_path = Path("./tests/tmp")
    temp_path.mkdir(parents=True, exist_ok=True)

    for file in temp_path.iterdir():
        file_path = temp_path / file
        if file_path.is_file() or file_path.is_symlink():
            file_path.unlink() 
        elif file_path.is_dir():
            shutil.rmtree(file_path) 
                    
    yield temp_path

@pytest.fixture
def trained_binary_model(classification_data, tmpdir):
    X, y = classification_data
    data = pd.concat([X, y], axis=1)
    trainer = TrainerSupervised(task='binary', output_dir=str(tmpdir))
    trainer.run(data=data, target_variable='target')
    return trainer

@pytest.fixture
def explainer_instance(trained_binary_model, tmpdir):
    trainer = trained_binary_model
    X_train, X_test = trainer.X_train, trainer.X_test
    y_test = trainer.y_test
    return Explainer(trainer, X_train, X_test, y_test, output_dir=tmpdir)

def test_explainer_initialization(explainer_instance, tmpdir):
    explainer = explainer_instance
    assert explainer.trainer is not None
    assert explainer.predictor is not None
    assert explainer.output_dir == tmpdir
    assert hasattr(explainer, 'X_train')
    assert hasattr(explainer, 'X_test')
    assert hasattr(explainer, 'y_test')

def test_explainer_run_binary_classification(explainer_instance):
    explainer = explainer_instance
    explainer.run()
    # Check if diagnostic plots are saved
    assert (explainer.output_dir / 'figures' / 'confusion_matrix.png').exists()
    assert (explainer.output_dir / 'figures' / 'feature_importance.png').exists()
    assert (explainer.output_dir / 'figures' / 'model_evaluation.png').exists()
    assert (explainer.output_dir / 'figures' / 'shap_heatmap.png').exists()
    assert (explainer.output_dir / 'figures' / 'shap_barplot.png').exists()

def test_explainer_from_trainer(trained_binary_model, tmpdir):
    trainer = trained_binary_model
    explainer = Explainer.from_trainer(trainer)
    assert explainer.trainer is trainer
    assert explainer.output_dir == trainer.output_dir
    assert explainer.X_train is trainer.X_train
    assert explainer.X_test is trainer.X_test
    assert explainer.y_test is trainer.y_test

@pytest.fixture
def trained_regression_model(regression_data, tmpdir):
    X, y = regression_data
    data = pd.concat([X, y], axis=1)
    trainer = TrainerSupervised(task='regression', output_dir=str(tmpdir))
    trainer.run(data=data, target_variable='target')
    return trainer 

def test_explainer_run_regression(trained_regression_model, tmpdir):
    trainer = trained_regression_model
    explainer = Explainer.from_trainer(trainer)
    explainer.run()
    # Check if regression diagnostic plots are saved
    assert (explainer.output_dir / 'figures' / 'residual_plot.png').exists()
    assert (explainer.output_dir / 'figures' / 'true_vs_predicted.png').exists()
    assert (explainer.output_dir / 'figures' / 'feature_importance.png').exists()




# ------------------------------------------------------------------------------
# Code adapted from:
# https://github.com/czifan/DeepSurv.pytorch
# ------------------------------------------------------------------------------

from typing import List

import numpy as np

import torch
import torch.nn as nn
import torch.optim as optim

import lightning.pytorch as pl

from .utils import calculate_c_index

class Regularization(object):
    def __init__(self, order: int, weight_decay: float):
        ''' The initialization of Regularization class

        :param order: (int) norm order number
        :param weight_decay: (float) weight decay rate
        '''
        super(Regularization, self).__init__()
        self.order = order
        self.weight_decay = weight_decay

    def __call__(self, model: torch.nn.Module):
        ''' Performs calculates regularization(self.order) loss for model.

        :param model: (torch.nn.Module object)
        :return reg_loss: (torch.Tensor) the regularization(self.order) loss
        '''
        reg_loss = 0
        for name, w in model.named_parameters():
            if 'weight' in name:
                reg_loss = reg_loss + torch.norm(w, p=self.order)
        reg_loss = self.weight_decay * reg_loss
        return reg_loss

class DeepSurv(nn.Module):
    ''' The module class performs building network according to config'''
    def __init__(self, in_channel: int, dims: List[int], dropout: float, norm: bool = True):
        
        super(DeepSurv, self).__init__()
        self.dropout = dropout
        self.norm = norm
        self.dims = [in_channel] + dims + [1]
        self.model = self._build_network()

    def _build_network(self):
        ''' Performs building networks according to parameters'''
        layers = []
        for i in range(len(self.dims)-1):
            if i and self.dropout is not None: # adds dropout layer
                layers.append(nn.Dropout(self.dropout))
            # adds linear layer
            layers.append(nn.Linear(self.dims[i], self.dims[i+1]))
            if self.norm: # adds batchnormalize layer
                layers.append(nn.BatchNorm1d(self.dims[i+1]))
            # adds activation layer
            layers.append(nn.ReLU())
        # builds sequential network
        return nn.Sequential(*layers)

    def forward(self, X):
        return self.model(X)

class NegativeLogLikelihood(nn.Module):
    def __init__(self, l2_reg):
        super(NegativeLogLikelihood, self).__init__()
        self.L2_reg = l2_reg
        self.reg = Regularization(order=2, weight_decay=self.L2_reg)

    def forward(self, risk_pred, y, e, model):
        mask = torch.ones(y.shape[0], y.shape[0])
        mask[(y.T - y) > 0] = 0
        log_loss = torch.exp(risk_pred) * mask
        log_loss = torch.sum(log_loss, dim=0) / torch.sum(mask, dim=0)
        log_loss = torch.log(log_loss).reshape(-1, 1)
        neg_log_loss = -torch.sum((risk_pred-log_loss) * e) / torch.sum(e)
        l2_loss = self.reg(model)
        return neg_log_loss + l2_loss
    
class LitDeepSurv(pl.LightningModule):
    def __init__(self, in_channel: int, dims: List[int], dropout: float, l2_reg: float):
        super(LitDeepSurv, self).__init__()
        self.save_hyperparameters() 
        
        self.model = DeepSurv(in_channel, dims=dims, dropout=dropout)
        self.criterion = NegativeLogLikelihood(l2_reg)
        self.best_c_index = 0

    def forward(self, x):
        return self.model(x)
    
    def predict(self, x):
        self.model.eval()
        
        with torch.no_grad():
            x = x.drop(columns=['time', 'event'], errors='ignore').to_numpy(dtype=np.float32)
            
            # Normalize input
            min_vals = x.min(axis=0)
            range_vals = x.max(axis=0) - min_vals
            range_vals[range_vals == 0] = 1  # Prevent division by zero
            
            x = (x - min_vals) / range_vals
            x = torch.from_numpy(x)
            
            y_pred = self.model(x).ravel()

        return y_pred

    def configure_optimizers(self) -> optim.Optimizer:
        return optim.Adam(self.model.parameters())

    def training_step(self, batch, _):
        X, y, e = batch
        risk_pred = self(X)
        train_loss = self.criterion(risk_pred, y, e, self.model)
        train_c = calculate_c_index(-risk_pred, y, e)

        self.log('train_loss', train_loss, prog_bar=True)
        self.log('train_c_index', train_c, prog_bar=True)
        return train_loss

    def validation_step(self, batch, _):
        X, y, e = batch
        risk_pred = self(X)
        valid_loss = self.criterion(risk_pred, y, e, self.model)
        valid_c = calculate_c_index(-risk_pred, y, e)

        self.log('valid_loss', valid_loss, prog_bar=True)
        self.log('valid_c_index', valid_c, prog_bar=True)

        return valid_loss


"""
Code adapted from https://github.com/epic-open-source/seismometer
"""

from pathlib import Path

import numpy as np
import matplotlib.pyplot as plt

from sklearn.calibration import calibration_curve
from sklearn.metrics import (
    precision_recall_curve,
    roc_auc_score,
    roc_curve,
)

from .functional import auprc, bootstrap_metric
import seaborn as sns

def _bin_class_curve(y_true: np.ndarray, y_pred: np.ndarray):
    sort_ix = np.argsort(y_pred, kind="mergesort")[::-1]
    y_true = np.array(y_true)[sort_ix]
    y_pred = np.array(y_pred)[sort_ix]

    # Find where the threshold changes
    distinct_ix = np.where(np.diff(y_pred))[0]
    threshold_idxs = np.r_[distinct_ix, y_true.size - 1]

    # Add up the true positives and infer false ones
    tps = np.cumsum(y_true)[threshold_idxs]
    fps = 1 + threshold_idxs - tps

    return fps, tps, y_pred[threshold_idxs]

def plot_epic_copy(
        y_test: np.ndarray,
        y_pred: np.ndarray,
        y_val: np.ndarray,
        y_val_pred: np.ndarray,
        y_train: np.ndarray,
        y_train_pred: np.ndarray,
        output_dir: Path
    ) -> None:

    # Compute test metrics
    fpr_test, tpr_test, thresholds_roc_test = roc_curve(y_test, y_pred)
    roc_auc_test = roc_auc_score(y_test, y_pred)
    precision_test, recall_test, thresholds_pr_test = precision_recall_curve(y_test, y_pred)
    average_precision_test = auprc(y_test, y_pred)
    prob_true_test, prob_pred_test = calibration_curve(y_test, y_pred, n_bins=10, strategy='uniform')

    # Compute validation metrics
    fpr_val, tpr_val, thresholds_roc_val = roc_curve(y_val, y_val_pred)
    roc_auc_val = roc_auc_score(y_val, y_val_pred)
    precision_val, recall_val, thresholds_pr_val = precision_recall_curve(y_val, y_val_pred)
    average_precision_val = auprc(y_val, y_val_pred)
    prob_true_val, prob_pred_val = calibration_curve(y_val, y_val_pred, n_bins=10, strategy='uniform')

    # Compute train metrics
    fpr_train, tpr_train, thresholds_roc_train = roc_curve(y_train, y_train_pred)
    roc_auc_train = roc_auc_score(y_train, y_train_pred)
    precision_train, recall_train, thresholds_pr_train = precision_recall_curve(y_train, y_train_pred)
    average_precision_train = auprc(y_train, y_train_pred)
    prob_true_train, prob_pred_train = calibration_curve(y_train, y_train_pred, n_bins=10, strategy='uniform')

    # Compute confidence intervals
    roc_conf_test = [round(val, 2) for val in np.percentile(bootstrap_metric(y_test, y_pred, roc_auc_score), (2.5, 97.5))]
    roc_conf_val = [round(val, 2) for val in np.percentile(bootstrap_metric(y_val, y_val_pred, roc_auc_score), (2.5, 97.5))]
    roc_conf_train = [round(val, 2) for val in np.percentile(bootstrap_metric(y_train, y_train_pred, roc_auc_score), (2.5, 97.5))]

    precision_conf_test = [round(val, 2) for val in np.percentile(bootstrap_metric(y_test, y_pred, auprc), (2.5, 97.5))]
    precision_conf_val = [round(val, 2) for val in np.percentile(bootstrap_metric(y_val, y_val_pred, auprc), (2.5, 97.5))]
    precision_conf_train = [round(val, 2) for val in np.percentile(bootstrap_metric(y_train, y_train_pred, auprc), (2.5, 97.5))]

    # Set Seaborn style
    sns.set_theme(style="darkgrid", font="Arial")

    plt.figure(figsize=(37.5, 10))

    # 1. ROC Curve
    plt.subplot(2, 5, 1)
    sns.lineplot(x=fpr_test, y=tpr_test, label=f"Test AUROC = {roc_auc_test:.2f} {roc_conf_test}", color="blue")
    sns.lineplot(x=fpr_val, y=tpr_val, label=f"Validation AUROC = {roc_auc_val:.2f} {roc_conf_val}", color="orange")
    sns.lineplot(x=fpr_train, y=tpr_train, label=f"Train AUROC = {roc_auc_train:.2f} {roc_conf_train}", color="green")
    plt.xlabel("1 - Specificity")
    plt.ylabel("Sensitivity")
    plt.title("ROC Curve")
    plt.legend()

    # 2. Precision-Recall Curve
    plt.subplot(2, 5, 2)
    sns.lineplot(x=recall_test, y=precision_test, label=f"Test AUC-PR = {average_precision_test:.2f} {precision_conf_test}", color="blue")
    sns.lineplot(x=recall_val, y=precision_val, label=f"Validation AUC-PR = {average_precision_val:.2f} {precision_conf_val}", color="orange")
    sns.lineplot(x=recall_train, y=precision_train, label=f"Train AUC-PR = {average_precision_train:.2f} {precision_conf_train}", color="green")
    plt.xlabel("Recall")
    plt.ylabel("Precision")
    plt.title("Precision-Recall Curve")
    plt.legend()

    # 3. Calibration Curve
    plt.subplot(2, 5, 6)
    sns.lineplot(x=prob_pred_test, y=prob_true_test, label="Test Calibration Curve", color="blue", marker='o')
    sns.lineplot(x=prob_pred_val, y=prob_true_val, label="Validation Calibration Curve", color="orange", marker='o')
    sns.lineplot(x=prob_pred_train, y=prob_true_train, label="Train Calibration Curve", color="green", marker='o')
    sns.lineplot(x=[0, 1], y=[0, 1], linestyle="--", label="Perfect Calibration", color="gray")
    plt.xlabel("Predicted Probability")
    plt.ylabel("Observed Probability")
    plt.title("Calibration Curve")
    plt.legend()

    # 4. Sensitivity vs Flag Rate
    fps, tps, _ = _bin_class_curve(y_test, y_pred)
    sens_test = tps / sum(y_test)
    flag_rate_test = (tps + fps) / len(y_test)

    fps, tps, _ = _bin_class_curve(y_val, y_val_pred)
    sens_val = tps / sum(y_val)
    flag_rate_val = (tps + fps) / len(y_val)

    fps, tps, _ = _bin_class_curve(y_train, y_train_pred)
    sens_train = tps / sum(y_train)
    flag_rate_train = (tps + fps) / len(y_train)

    plt.subplot(2, 5, 7)
    sns.lineplot(x=flag_rate_test, y=sens_test, label="Test", color="blue")
    sns.lineplot(x=flag_rate_val, y=sens_val, label="Validation", color="orange")
    sns.lineplot(x=flag_rate_train, y=sens_train, label="Train", color="green")
    plt.xlabel('Flag Rate')
    plt.ylabel('Sensitivity')
    plt.title('Sensitivity/Flag Curve')
    plt.legend()

    # 5. Sensitivity, Specificity, PPV by Threshold
    # Test Metrics
    sensitivity_test = tpr_test
    specificity_test = 1 - fpr_test
    ppv_test = precision_test[:-1]

    # Validation Metrics
    sensitivity_val = tpr_val
    specificity_val = 1 - fpr_val
    ppv_val = precision_val[:-1]

    # Train Metrics
    sensitivity_train = tpr_train
    specificity_train = 1 - fpr_train
    ppv_train = precision_train[:-1]

    # Plot Test Metrics
    plt.subplot(2, 5, 3)
    sns.lineplot(x=thresholds_roc_test, y=sensitivity_test, label="Test Sensitivity", color="blue")
    sns.lineplot(x=thresholds_roc_test, y=specificity_test, label="Test Specificity", color="green")
    sns.lineplot(x=thresholds_pr_test, y=ppv_test, label="Test PPV", color="magenta")
    plt.xlabel("Threshold")
    plt.ylabel("Metric")
    plt.title("Test Metrics by Threshold")
    plt.legend()

    # Plot Validation Metrics
    plt.subplot(2, 5, 4)
    sns.lineplot(x=thresholds_roc_val, y=sensitivity_val, label="Validation Sensitivity", linestyle="--", color="orange")
    sns.lineplot(x=thresholds_roc_val, y=specificity_val, label="Validation Specificity", linestyle="--", color="darkgreen")
    sns.lineplot(x=thresholds_pr_val, y=ppv_val, label="Validation PPV", linestyle="--", color="pink")
    plt.xlabel("Threshold")
    plt.ylabel("Metric")
    plt.title("Validation Metrics by Threshold")
    plt.legend()

    # Plot Train Metrics
    plt.subplot(2, 5, 5)
    sns.lineplot(x=thresholds_roc_train, y=sensitivity_train, label="Train Sensitivity", linestyle=":", color="purple")
    sns.lineplot(x=thresholds_roc_train, y=specificity_train, label="Train Specificity", linestyle=":", color="brown")
    sns.lineplot(x=thresholds_pr_train, y=ppv_train, label="Train PPV", linestyle=":", color="cyan")
    plt.xlabel("Threshold")
    plt.ylabel("Metric")
    plt.title("Train Metrics by Threshold")
    plt.legend()

    # 6. Histogram of Predicted Probabilities
    def _get_highest_bin_count(values, bins):
        counts, _ = np.histogram(values, bins=bins)
        return counts.max()

    highest_bin_count = max(
        _get_highest_bin_count(y_pred[y_test == 0], bins=20),
        _get_highest_bin_count(y_pred[y_test == 1], bins=20),
        _get_highest_bin_count(y_val_pred[y_val == 0], bins=20),
        _get_highest_bin_count(y_val_pred[y_val == 1], bins=20),
        _get_highest_bin_count(y_train_pred[y_train == 0], bins=20),
        _get_highest_bin_count(y_train_pred[y_train == 1], bins=20)
    )
    highest_bin_count += highest_bin_count//20

    plt.subplot(2, 5, 8)
    sns.histplot(y_pred[y_test == 0], bins=20, alpha=0.7, label="Test Actual False", color='blue', kde=False)
    sns.histplot(y_pred[y_test == 1], bins=20, alpha=0.7, label="Test Actual True", color='magenta', kde=False)
    plt.xlabel("Predicted Probability")
    plt.ylabel("Count")
    plt.title("Histogram of Predicted Probabilities")
    plt.ylim(0, highest_bin_count)
    plt.legend()

    plt.subplot(2, 5, 9)
    sns.histplot(y_val_pred[y_val == 0], bins=20, alpha=0.5, label="Validation Actual False", color='orange', kde=False)
    sns.histplot(y_val_pred[y_val == 1], bins=20, alpha=0.5, label="Validation Actual True", color='pink', kde=False)
    plt.xlabel("Predicted Probability")
    plt.ylabel("Count")
    plt.title("Histogram of Predicted Probabilities")
    plt.ylim(0, highest_bin_count)
    plt.legend()

    plt.subplot(2, 5, 10)
    sns.histplot(y_train_pred[y_train == 0], bins=20, alpha=0.5, label="Train Actual False", color='green', kde=False)
    sns.histplot(y_train_pred[y_train == 1], bins=20, alpha=0.5, label="Train Actual True", color='purple', kde=False)
    plt.xlabel("Predicted Probability")
    plt.ylabel("Count")
    plt.title("Histogram of Predicted Probabilities")
    plt.ylim(0, highest_bin_count)
    plt.legend()

    plt.tight_layout()
    plt.savefig(output_dir / 'model_evaluation.png')
    plt.close()



from jarvais.analyzer import Analyzer
from pathlib import Path
import pytest
import numpy as np
import pandas as pd
import shutil

@pytest.fixture
def sample_data():
    data = {
        'A': [1, 2, 3, 4, 5],
        'B': [5, 4, 3, 2, 1],
        'C': ['a', 'b', 'a', 'a', 'b'],
        'D': [None, 2, None, 4, 5]
    }
    df= pd.concat([pd.DataFrame(data), pd.DataFrame(data), pd.DataFrame(data)], axis=0)
    return df.reset_index()

@pytest.fixture
def tmpdir():
    temp_path = Path("./tests/tmp")
    temp_path.mkdir(parents=True, exist_ok=True)

    for file in temp_path.iterdir():
        file_path = temp_path / file
        if file_path.is_file() or file_path.is_symlink():
            file_path.unlink() 
        elif file_path.is_dir():
            shutil.rmtree(file_path) 
                    
    yield temp_path

@pytest.fixture
def analyzer(sample_data, tmpdir):
    config_file = tmpdir / 'config.yaml'
    output_dir = tmpdir
    if "data.csv" not in output_dir.iterdir():
        sample_data.to_csv(output_dir / 'data.csv', index=False)
    return Analyzer(data=sample_data, output_dir=output_dir)

def test_analyzer_initialization(analyzer, sample_data):
    assert analyzer.data.equals(sample_data)
    assert analyzer.target_variable is None
    assert analyzer.output_dir is not None
    assert analyzer.config is None

def test_replace_missing(analyzer):
    analyzer._create_config()
    analyzer.config['missingness_strategy']['continuous']['D'] = 'mean'
    analyzer._apply_config() # replace missing happens in here
    
    assert analyzer.data['D'].isna().sum() == 0
    assert np.isclose(analyzer.data['D'].iloc[0], analyzer.data['D'].mean(), rtol=1e-4)

def test_create_config(analyzer):
    # _infer_types is always run inside _create_config
    analyzer._create_config()
    assert 'A' in analyzer.continuous_columns
    assert 'B' in analyzer.continuous_columns
    assert 'C' in analyzer.categorical_columns

def test_create_multiplots(analyzer):
    analyzer.categorical_columns = ['C']
    analyzer.continuous_columns = ['A', 'B']
    analyzer.umap_data = pd.DataFrame.from_dict({'UMAP1': [i for i in range(1, 16)], 'UMAP2': [j for j in range(15, 0, -1)]}).to_numpy()
    analyzer._create_multiplots(figures_dir=analyzer.output_dir / 'figures')
    assert len(analyzer.multiplots) > 0

def test_run(analyzer):
    analyzer.run()
    assert (analyzer.output_dir / 'tableone.csv').exists()
    assert (analyzer.output_dir / 'updated_data.csv').exists()
    assert (analyzer.output_dir / 'figures' / 'pearson_correlation.png').exists()
    assert (analyzer.output_dir / 'figures' / 'spearman_correlation.png').exists()
    assert (analyzer.output_dir / 'figures' / 'multiplots').exists()
    assert (analyzer.output_dir / 'figures' / 'frequency_tables').exists()



from typing import Union, List

import numpy as np
import pandas as pd
from scipy.interpolate import interp1d

import torch
from torch import nn
from torch import optim

import lightning.pytorch as pl

from .utils import make_optimizer, normalize

class MTLR(nn.Module):
    """Multi-task logistic regression for individualised
    survival prediction.

    The MTLR time-logits are computed as:
    `z = sum_k x^T w_k + b_k`,
    where `w_k` and `b_k` are learnable weights and biases for each time
    interval.

    Note that a slightly more efficient reformulation is used here, first
    proposed in [2]_.

    References
    ----------
    ..[1] C.-N. Yu et al., ‘Learning patient-specific cancer survival
    distributions as a sequence of dependent regressors’, in Advances in neural
    information processing systems 24, 2011, pp. 1845–1853.
    ..[2] P. Jin, ‘Using Survival Prediction Techniques to Learn
    Consumer-Specific Reservation Price Distributions’, Master's thesis,
    University of Alberta, Edmonton, AB, 2015.
    """

    def __init__(self, in_features: int, num_time_bins: int):
        """Initialises the module.

        Parameters
        ----------
        in_features
            Number of input features.
        num_time_bins
            The number of bins to divide the time axis into.
        """
        super().__init__()
        if num_time_bins < 1:
            raise ValueError("The number of time bins must be at least 1")
        if in_features < 1:
            raise ValueError("The number of input features must be at least 1")
        self.in_features = in_features
        self.num_time_bins = num_time_bins + 1 # + extra time bin [max_time, inf)

        self.mtlr_weight = nn.Parameter(torch.Tensor(self.in_features,
                                                     self.num_time_bins - 1))
        self.mtlr_bias = nn.Parameter(torch.Tensor(self.num_time_bins - 1))

        # `G` is the coding matrix from [2]_ used for fast summation.
        # When registered as buffer, it will be automatically
        # moved to the correct device and stored in saved
        # model state.
        self.register_buffer(
            "G",
            torch.tril(
                torch.ones(self.num_time_bins - 1,
                           self.num_time_bins,
                           requires_grad=True)))
        self.reset_parameters()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Performs a forward pass on a batch of examples.

        Parameters
        ----------
        x : torch.Tensor, shape (num_samples, num_features)
            The input data.

        Returns
        -------
        torch.Tensor, shape (num_samples, num_time_bins)
            The predicted time logits.
        """
        out = torch.matmul(x, self.mtlr_weight) + self.mtlr_bias
        return torch.matmul(out, self.G)

    def reset_parameters(self):
        """Resets the model parameters."""
        nn.init.xavier_normal_(self.mtlr_weight)
        nn.init.constant_(self.mtlr_bias, 0.)

    def __repr__(self):
        return (f"{self.__class__.__name__}(in_features={self.in_features},"
                f" num_time_bins={self.num_time_bins})")

class LitMTLR(pl.LightningModule):
    def __init__(
            self, 
            in_channel: int, 
            num_time_bins: int, 
            dims: List[int], 
            dropout: float, 
            C1: float,
            mean: pd.Series | None = None,
            std: pd.Series | None = None,
        ):
        super(LitMTLR, self).__init__()
        self.save_hyperparameters() 

        self.in_channel = in_channel
        self.num_time_bins = num_time_bins
        self.dims = [in_channel] + dims
        self.dropout = dropout
        self.C1 = C1

        self.training_mean = mean.drop(['time', 'event']) if not mean is None else mean
        self.training_std = std.drop(['time', 'event']) if not std is None else std

        self._build_model()

    def _build_model(self):

        layers = []
        for i in range(len(self.dims)-1):
            layers.append(nn.Linear(self.dims[i], self.dims[i+1]))
            layers.append(nn.ELU())
            layers.append(nn.Dropout(self.dropout))
            layers.append(nn.BatchNorm1d(self.dims[i+1]))
        layers.append(MTLR(in_features=self.dims[-1], num_time_bins=self.num_time_bins))
        self.model = nn.Sequential(*layers)

    def forward(self, x):
        return self.model(x)
    
    def predict(self, x: pd.DataFrame):
        x = x.drop(["time", "event"], axis=1, errors='ignore')
        skip_cols = [col for col in x.columns if (set(x[col].unique()).issubset({0, 1}))]
        x, _, _ = normalize(x, mean=self.training_mean, std=self.training_std, skip_cols=skip_cols)

        self.model.eval()
        with torch.no_grad():
            y_pred = self.model(torch.tensor(x.values, dtype=torch.float))
            risk_pred = mtlr_risk(y_pred)

        return risk_pred

    def configure_optimizers(self) -> optim.Optimizer:
        return make_optimizer(optim.Adam, self.model, lr=.005, weight_decay=1e-5)

    def training_step(self, batch, _):
        X, y = batch
        pred = self(X)
        train_loss = mtlr_neg_log_likelihood(pred, y, self.model, C1=self.C1, average=True)

        self.log('train_loss', train_loss, prog_bar=True)
        return train_loss

    def validation_step(self, batch, _):
        X, y = batch
        pred = self(X)
        valid_loss = mtlr_neg_log_likelihood(pred, y, self.model, C1=self.C1, average=True)

        self.log('valid_loss', valid_loss, prog_bar=True)
        return valid_loss

def masked_logsumexp(x: torch.Tensor,
                     mask: torch.Tensor,
                     dim: int = -1) -> torch.Tensor:
    """Computes logsumexp over elements of a tensor specified by a mask
    in a numerically stable way.

    Parameters
    ----------
    x
        The input tensor.
    mask
        A tensor with the same shape as `x` with 1s in positions that should
        be used for logsumexp computation and 0s everywhere else.
    dim
        The dimension of `x` over which logsumexp is computed. Default -1 uses
        the last dimension.

    Returns
    -------
    torch.Tensor
        Tensor containing the logsumexp of each row of `x` over `dim`.
    """
    max_val, _ = (x * mask).max(dim=dim)
    max_val = torch.clamp_min(max_val, 0)
    return torch.log(
        torch.sum(torch.exp(x - max_val.unsqueeze(dim)) * mask,
                  dim=dim)) + max_val


def mtlr_neg_log_likelihood(logits: torch.Tensor,
                            target: torch.Tensor,
                            model: torch.nn.Module,
                            C1: float,
                            average: bool = False) -> torch.Tensor:
    """Computes the negative log-likelihood of a batch of model predictions.

    Parameters
    ----------
    logits : torch.Tensor, shape (num_samples, num_time_bins)
        Tensor with the time-logits (as returned by the MTLR module) for one
        instance in each row.
    target : torch.Tensor, shape (num_samples, num_time_bins)
        Tensor with the encoded ground truth survival.
    model
        PyTorch Module with at least `MTLR` layer.
    C1
        The L2 regularization strength.
    average
        Whether to compute the average log likelihood instead of sum
        (useful for minibatch training).

    Returns
    -------
    torch.Tensor
        The negative log likelihood.
    """
    censored = target.sum(dim=1) > 1
    nll_censored = masked_logsumexp(logits[censored], target[censored]).sum() if censored.any() else 0
    nll_uncensored = (logits[~censored] * target[~censored]).sum() if (~censored).any() else 0

    # the normalising constant
    norm = torch.logsumexp(logits, dim=1).sum()

    nll_total = -(nll_censored + nll_uncensored - norm)
    if average:
        nll_total = nll_total / target.size(0)

    # L2 regularization
    for k, v in model.named_parameters():
        if "mtlr_weight" in k:
            nll_total += C1/2 * torch.sum(v**2)

    return nll_total


def mtlr_survival(logits: torch.Tensor) -> torch.Tensor:
    """Generates predicted survival curves from predicted logits.

    Parameters
    ----------
    logits
        Tensor with the time-logits (as returned by the MTLR module) for one
        instance in each row.

    Returns
    -------
    torch.Tensor
        The predicted survival curves for each row in `pred` at timepoints used
        during training.
    """
    # TODO: do not reallocate G in every call
    G = torch.tril(torch.ones(logits.size(1),
                              logits.size(1))).to(logits.device)
    density = torch.softmax(logits, dim=1)
    return torch.matmul(density, G)


def mtlr_survival_at_times(logits: torch.Tensor,
                           train_times: Union[torch.Tensor, np.ndarray],
                           pred_times: np.ndarray) -> np.ndarray:
    """Generates predicted survival curves at arbitrary timepoints using linear
    interpolation.

    Notes
    -----
    This function uses scipy.interpolate internally and returns a Numpy array,
    in contrast with `mtlr_survival`.

    Parameters
    ----------
    logits
        Tensor with the time-logits (as returned by the MTLR module) for one
        instance in each row.
    train_times
        Time bins used for model training. Must have the same length as the
        first dimension of `pred`.
    pred_times
        Array of times used to compute the survival curve.

    Returns
    -------
    np.ndarray
        The survival curve for each row in `pred` at `pred_times`. The values
        are linearly interpolated at timepoints not used for training.
    """
    train_times = np.pad(train_times, (1, 0))
    surv = mtlr_survival(logits).detach().cpu().numpy()
    interpolator = interp1d(train_times, surv)
    return interpolator(np.clip(pred_times, 0, train_times.max()))


def mtlr_hazard(logits: torch.Tensor) -> torch.Tensor:
    """Computes the hazard function from MTLR predictions.

    The hazard function is the instantenous rate of failure, i.e. roughly
    the risk of event at each time interval. It's computed using
    `h(t) = f(t) / S(t)`,
    where `f(t)` and `S(t)` are the density and survival functions at t,
    respectively.

    Parameters
    ----------
    logits
        The predicted logits as returned by the `MTLR` module.

    Returns
    -------
    torch.Tensor
        The hazard function at each time interval in `y_pred`.
    """
    return torch.softmax(
        logits, dim=1)[:, :-1] / (mtlr_survival(logits) + 1e-15)[:, 1:]


def mtlr_risk(logits: torch.Tensor) -> torch.Tensor:
    """Computes the overall risk of event from MTLR predictions.

    The risk is computed as the time integral of the cumulative hazard,
    as defined in [1]_.

    Parameters
    ----------
    logits
        The predicted logits as returned by the `MTLR` module.

    Returns
    -------
    torch.Tensor
        The predicted overall risk.
    """
    hazard = mtlr_hazard(logits)
    return torch.sum(hazard.cumsum(1), dim=1)



import logging
from pathlib import Path

import pandas as pd
import torch
from torch.utils.data import DataLoader

import lightning.pytorch as pl
import optuna
from optuna.integration import PyTorchLightningPruningCallback

from .deepsurv import LitDeepSurv
from .utils import SurvivalDataset
from .utils import calculate_c_index

# Suppress PyTorch Lightning logging
logging.getLogger("lightning.pytorch.utilities.rank_zero").setLevel(logging.FATAL)

def train_deepsurv(data_train: pd.DataFrame, data_val:pd.DataFrame, data_test: pd.DataFrame, output_dir: Path):

    in_channel = len(data_train.columns) - 2 # -2 to remove time and event

    # Initialize datasets and dataloaders
    train_dataset = SurvivalDataset(data_train)
    val_dataset = SurvivalDataset(data_val)
    test_dataset = SurvivalDataset(data_test)

    train_loader = DataLoader(train_dataset, batch_size=len(train_dataset))
    val_loader = DataLoader(val_dataset, batch_size=len(val_dataset))
    test_loader = DataLoader(test_dataset, batch_size=len(test_dataset))

    def objective(trial: optuna.trial.Trial) -> float:
        l2_reg = trial.suggest_float("l2_reg", 1e-6, 1e-2)
        dropout = trial.suggest_float("dropout", 0.2, 0.5)
        dims = trial.suggest_categorical("dims", [[2**n, 2**n, 2**n] for n in range(4, 9)])

        model = LitDeepSurv(in_channel=in_channel, dims=dims, dropout=dropout, l2_reg=l2_reg)

        trainer = pl.Trainer(
            enable_checkpointing=False,
            enable_progress_bar=False,
            enable_model_summary=False,
            max_epochs=10,
            accelerator="auto",
            callbacks=[PyTorchLightningPruningCallback(trial, monitor="valid_c_index")],
        )
        trainer.fit(model, train_loader, val_loader)

        return trainer.callback_metrics["valid_c_index"].item()
    
    pruner = optuna.pruners.MedianPruner() 

    optuna.logging.set_verbosity(optuna.logging.WARNING)
    study = optuna.create_study(direction="maximize", pruner=pruner)
    print('Training DeepSurv...')
    study.optimize(objective, n_trials=100, timeout=600)

    print("  Best trial:")
    trial = study.best_trial
    print("    Params: ")
    for key, value in trial.params.items():
        print("      {}: {}".format(key, value))

    model = LitDeepSurv(in_channel=in_channel, **trial.params)

    trainer = pl.Trainer(
        default_root_dir=output_dir,
        enable_checkpointing=False,
        enable_progress_bar=False,
        enable_model_summary=False,
        max_epochs=500,
        accelerator="auto",
        callbacks=[
            pl.callbacks.EarlyStopping(monitor="valid_c_index", mode="max")
        ]
    )
    trainer.fit(model, train_loader, val_loader)
    trainer.save_checkpoint(output_dir / 'DeepSurv.ckpt')

    # Prediction on test data
    model.eval()
    c_index = 0
    with torch.no_grad():
        for X, y, e in test_loader:
            risk_pred = model(X)  # Get risk predictions
            c_index += calculate_c_index(-risk_pred, y, e)
        c_index /= len(test_loader)
    
    return model, c_index


import pytest
import pandas as pd
import numpy as np
import shutil
from pathlib import Path
from sklearn.datasets import make_classification, make_regression
from jarvais.trainer import TrainerSupervised

@pytest.fixture
def classification_data():
    X, y = make_classification(
        n_samples=50, n_features=5, n_informative=3, n_redundant=1, random_state=42
    )
    X = pd.DataFrame(X, columns=[f"feature{i}" for i in range(1, X.shape[1] + 1)])
    y = pd.Series(y, name="target")
    return X, y

@pytest.fixture
def regression_data():
    X, y = make_regression(
        n_samples=50, n_features=5, noise=0.1, random_state=42
    )
    X = pd.DataFrame(X, columns=[f"feature{i}" for i in range(1, X.shape[1] + 1)])
    y = pd.Series(y, name="target")
    return X, y

@pytest.fixture
def survival_data():
    """Synthetic survival data generation."""
    n_samples = 50
    np.random.seed(42)
    X = np.random.normal(size=(n_samples, 5))  # Covariates
    event_time = np.random.exponential(scale=10, size=n_samples)  # Simulated survival times
    censor_time = np.random.uniform(0, 15, size=n_samples)  # Censoring times
    observed_time = np.minimum(event_time, censor_time)  # Observed time
    event_observed = (event_time <= censor_time).astype(int)  # Event indicator
    X_df = pd.DataFrame(X, columns=[f"feature{i}" for i in range(1, 6)])
    data = pd.concat([X_df, pd.Series(observed_time, name="time"), pd.Series(event_observed, name="event")], axis=1)
    return data

@pytest.fixture
def tmpdir():
    temp_path = Path("./tests/tmp")
    temp_path.mkdir(parents=True, exist_ok=True)

    for file in temp_path.iterdir():
        file_path = temp_path / file
        if file_path.is_file() or file_path.is_symlink():
            file_path.unlink()
        elif file_path.is_dir():
            shutil.rmtree(file_path)

    yield temp_path

def test_trainer_initialization():
    trainer = TrainerSupervised(task='binary')
    assert trainer.task == 'binary'
    assert trainer.reduction_method is None
    assert trainer.keep_k == 2
    with pytest.raises(ValueError):
        TrainerSupervised(task='invalid_task')

def test_feature_reduction(classification_data):
    X, y = classification_data
    trainer = TrainerSupervised(task='binary', reduction_method='variance_threshold')
    X_reduced = trainer._feature_reduction(X, y)
    assert X_reduced.shape[1] <= X.shape[1]

def test_run_method_classification(classification_data, tmpdir):
    X, y = classification_data
    data = pd.concat([X, y], axis=1)
    trainer = TrainerSupervised(task='binary', output_dir=str(tmpdir))
    trainer.run(data=data, target_variable='target')
    data_dir = tmpdir / 'data'
    assert (data_dir / 'X_train.csv').exists()
    assert (data_dir / 'X_test.csv').exists()
    assert hasattr(trainer, 'predictor')
    assert hasattr(trainer, 'X_train')
    assert hasattr(trainer, 'X_test')

def test_run_method_regression(regression_data, tmpdir):
    X, y = regression_data
    data = pd.concat([X, y], axis=1)
    trainer = TrainerSupervised(task='regression', output_dir=str(tmpdir))
    trainer.run(data=data, target_variable='target')
    data_dir = tmpdir / 'data'
    assert (data_dir / 'X_train.csv').exists()
    assert (data_dir / 'X_test.csv').exists()
    assert hasattr(trainer, 'predictor')
    assert hasattr(trainer, 'X_train')
    assert hasattr(trainer, 'X_test')

def test_run_method_survival(survival_data, tmpdir):
    data = survival_data
    trainer = TrainerSupervised(task='survival', output_dir=str(tmpdir))
    trainer.run(data=data, target_variable=["time", "event"])
    data_dir = tmpdir / 'data'
    assert (data_dir / 'X_train.csv').exists()
    assert (data_dir / 'X_test.csv').exists()
    assert hasattr(trainer, 'predictor')
    assert hasattr(trainer, 'X_train')
    assert hasattr(trainer, 'X_test')

def test_load_trainer_autogluon(classification_data, tmpdir):
    X, y = classification_data
    data = pd.concat([X, y], axis=1)
    trainer = TrainerSupervised(task='binary', output_dir=str(tmpdir))
    trainer.run(data=data, target_variable='target')
    loaded_trainer = TrainerSupervised.load_trainer(project_dir=str(tmpdir))
    assert loaded_trainer.predictor is not None

def test_load_trainer_survival(survival_data, tmpdir):
    data = survival_data
    trainer = TrainerSupervised(task='survival', output_dir=str(tmpdir))
    trainer.run(data=data, target_variable=["time", "event"])
    loaded_trainer = TrainerSupervised.load_trainer(project_dir=str(tmpdir))
    assert loaded_trainer.predictor is not None


import re
import inspect
from functools import partial

import pandas as pd
import fairlearn as fm

def infer_sensitive_features(data: pd.DataFrame) -> pd.DataFrame:
    """
    Infers potentially sensitive features from a DataFrame.
    """
    sensitive_keywords = [
        "gender", "sex", "age", "race", "ethnicity", "income",
        "religion", "disability", "nationality", "language",
        "marital", "citizen", "veteran", "status", "orientation",
        "disease", "regimen", "disease_site"
    ]

    sensitive_features = {
        col for col in data.columns 
        if any(re.search(rf"\b{keyword}\b", col, re.IGNORECASE) for keyword in sensitive_keywords)
    }

    return data[sensitive_features]

def _get_metric(metric, sensitive_features=None):
    fn = getattr(fm, metric)
    params = inspect.signature(fn).parameters
    return partial(fn, sensitive_features=sensitive_features) if 'sensitive_features' in params and sensitive_features else fn



from itertools import combinations
from typing import TYPE_CHECKING
from pathlib import Path

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns

import shap
from autogluon.tabular import TabularPredictor
from scipy.stats import f_oneway, ttest_ind
from sklearn.metrics import (
    confusion_matrix,
    r2_score,
    roc_auc_score,
    root_mean_squared_error,
)
from umap import UMAP

from lifelines.statistics import multivariate_logrank_test
from sksurv.nonparametric import kaplan_meier_estimator

from .functional import auprc, ci_wrapper, bootstrap_metric
from ._plot_epic import plot_epic_copy
from ._design import config_plot

# ANALYZER
@config_plot(plot_type='multi')
def plot_one_multiplot(
        data: pd.DataFrame,
        umap_data: pd.DataFrame,
        var: str,
        continuous_columns: list,
        output_dir: Path
    ) -> Path:

    def prep_for_pie(df, label):
        # Prepares data for pie plotting by grouping and sorting values.
        data = df.groupby(label).size().sort_values(ascending=False)

        labels = data.index.tolist()
        values = data.values.tolist()

        return labels, values

    # only write % if big enough
    def autopct(pct):
        return ('%1.1f%%' % pct) if pct > 3.5 else ''

    def calculate_fontsize(num_categories):
        base_fontsize = 16
        min_fontsize = 8
        return max(base_fontsize - num_categories * 1.5, min_fontsize)

    num_categories = len(data[var].unique())

    labels, values = prep_for_pie(data, var)

    fontsize = calculate_fontsize(num_categories)

    # setting number of rows/columns for subplots
    n = len(continuous_columns) + 2
    if n < 21:
        cols = 4
    elif n < 36:
        cols = 5
    elif n < 49:
        cols = 6
    elif n < 71:
        cols = 7
    elif n < 89: 
        cols = 8
    else: # i hope it never gets to this
        cols = 9

    rows = int(np.ceil(n / cols))  # int(np.ceil(np.sqrt(n)))
    scaler = 6

    # create subplot grid
    fig, ax = plt.subplots(rows, cols, figsize=((cols-0.5)*scaler, rows*scaler), dpi=300)
    ax = ax.flatten()

    # Pie plot of categorical variable
    ax[0].pie(values,
              labels=labels,
              autopct=autopct,
              startangle=90,   # 90 = 12 o'clock, 0 = 3 o'clock, 180 = 9 o'clock
              counterclock=False,
              textprops={'fontsize': fontsize},
              colors=plt.cm.Set2.colors)
    ax[0].set_title(f"{var} Distribution. N: {data[var].count()}")

    # UMAP colored by variable
    sns.scatterplot(x=umap_data[:,0], y=umap_data[:,1], hue=data[var], alpha=.7, ax=ax[1])
    ax[1].set_title(f'UMAP of Continuous Variables with {var}')
    if data[var].nunique() > 5: # Puts legend under plot if there are too many categories
        ax[1].legend(loc='upper center', bbox_to_anchor=(0.5, -0.15), ncol=3)

    p_values = {}

    # Calculate p-values
    for col in continuous_columns:
        unique_values = data[var].unique()

        if len(unique_values) > 1:
            if len(unique_values) == 2: # If binary classification, use t-test
                group1 = data[data[var] == unique_values[0]][col]
                group2 = data[data[var] == unique_values[1]][col]
                _, p_value = ttest_ind(group1, group2, equal_var=False)
            else: # For more than two categories, use ANOVA
                groups = [data[data[var] == value][col] for value in unique_values]
                _, p_value = f_oneway(*groups)
            p_values[col] = p_value
            sorted_columns = sorted(p_values, key=p_values.get)

            for i, col in enumerate(sorted_columns):
                sns.violinplot(x=var, y=col, data=data, ax=ax[i+2], inner="point")
                ax[i+2].tick_params(axis='x', labelrotation=67.5)
                ax[i+2].set_title(f"{var} vs {col} (p-value: {p_values[col]:.4f})")

    # Turn off unused axes
    for j in range(n, len(ax)):
        fig.delaxes(ax[j])

    multiplot_path = output_dir / 'multiplots' / f'{var}_multiplots.png'
    plt.savefig(multiplot_path)
    plt.close()

    return multiplot_path

@config_plot()
def plot_corr(
        corr: pd.DataFrame,
        size: float,
        output_dir: Path,
        file_name: str = 'correlation_matrix.png',
        title: str = "Correlation Matrix"
    ) -> None:
    """
    Plots a lower-triangle heatmap of the correlation matrix and saves it as an image file.

    Args:
        corr (pd.DataFrame): Correlation matrix to visualize.
        size (float): Size of the heatmap figure.
        output_dir (Path): Directory to save the output image.
        file_name (str): Name of the saved image file. Defaults to 'correlation_matrix.png'.

    Example:
        ```python
        import pandas as pd
        from pathlib import Path

        # Sample data
        df = pd.DataFrame({
            'A': [1, 2, 3, 4, 5],
            'B': [5, 4, 3, 2, 1],
            'C': [2, 3, 4, 5, 6]
        })

        # Compute Spearman correlation
        corr_matrix = df.corr(method='spearman')

        # Plot and save the heatmap
        plot_corr(corr=corr_matrix, size=6, output_dir=Path('./output'))
        ```
    """
    fig, ax = plt.subplots(1, 1, figsize=(size*1.2, size))
    mask = np.triu(np.ones_like(corr, dtype=bool)) # Keep only lower triangle
    np.fill_diagonal(mask, False)
    sns.heatmap(corr, mask=mask, annot=True, cmap='coolwarm', vmin=-1, vmax=1, linewidth=.5, fmt="1.2f", ax=ax)
    plt.title(title)
    plt.tight_layout()

    figure_path = output_dir / file_name
    fig.savefig(figure_path)
    plt.close()

@config_plot(plot_type='ft')
def plot_frequency_table(
        data: pd.DataFrame,
        columns: list,
        output_dir: Path
    ) -> None:
    """
    Generates and saves heatmap visualizations for frequency tables of all column pair combinations.

    Args:
        data (pd.DataFrame): Input dataset containing the columns to analyze.
        columns (list): List of column names to create frequency tables for.
        output_dir (Path): Directory to save the generated heatmaps.
    """
    frequency_dir = Path(output_dir) / 'frequency_tables'
    frequency_dir.mkdir(parents=True, exist_ok=True)

    for column_1, column_2 in combinations(columns, 2):
        heatmap_data = pd.crosstab(data[column_1], data[column_2])
        plt.figure(figsize=(8, 6))
        sns.heatmap(heatmap_data, annot=True, cmap='coolwarm', fmt='d', linewidth=.5)
        plt.title(f'Frequency Table for {column_1} and {column_2}')
        plt.xlabel(column_2)
        plt.ylabel(column_1)
        plt.savefig(frequency_dir / f'{column_1}_vs_{column_2}.png')
        plt.close()

@config_plot()
def plot_pairplot(
        data: pd.DataFrame,
        continuous_columns: list,
        output_dir: Path,
        target_variable: str = None,
        n_keep: int = 10
    ) -> None:
    """
    Generates a pair plot of the specified continuous columns in the dataset.

    Args:
        data (pd.DataFrame): The input DataFrame containing the data to be visualized.
        continuous_columns (list): A list of column names corresponding to continuous variables.
        output_dir (Path): Directory where the resulting plot will be saved.
        target_variable (str, optional): The target variable to use as a hue for coloring the pair plot. Defaults to None.
        n_keep (int, optional): The maximum number of continuous columns to include in the plot. 
            If exceeded, the most correlated columns are selected. Defaults to 10.
    """
    if len(continuous_columns) > n_keep:
        spearman_corr = data[continuous_columns].corr(method="spearman") 
        corr_pairs = spearman_corr.abs().unstack().sort_values(
            kind="quicksort",
            ascending=False
        ).drop_duplicates()
        top_10_pairs = corr_pairs[corr_pairs < 1].nlargest(5)
        columns_to_plot = list({index for pair in top_10_pairs.index for index in pair})
    else:
        columns_to_plot = continuous_columns

    hue = target_variable
    if target_variable is not None:
        columns_to_plot += [target_variable]

    sns.set_theme(style="darkgrid", font="Arial")
    g = sns.pairplot(data[columns_to_plot], hue=hue)
    g.figure.suptitle("Pair Plot", y=1.08)

    figure_path = output_dir / 'pairplot.png'
    plt.savefig(figure_path)
    plt.close()

@config_plot()
def plot_umap(
        data: pd.DataFrame,
        continuous_columns: list,
        output_dir: Path
    ) -> np.ndarray:
    """
    Generates a 2D UMAP projection of the specified continuous columns and saves the resulting scatter plot.

    Args:
        data (pd.DataFrame): The input DataFrame containing the data to be visualized.
        continuous_columns (list): A list of column names corresponding to continuous variables 
            to be included in the UMAP projection.
        output_dir (Path): Directory where the resulting plot will be saved.

    Returns:
        np.ndarray: A 2D NumPy array of the UMAP-transformed data.
    """
    umap_data = UMAP(n_components=2).fit_transform(data[continuous_columns])

    plt.figure(figsize=(8, 8))
    sns.scatterplot(x=umap_data[:,0], y=umap_data[:,1], alpha=.7)
    plt.title('UMAP of Continuous Variables')
    plt.savefig(output_dir / 'umap_continuous_data.png')
    plt.close()

    return umap_data

@config_plot()
def plot_kaplan_meier_by_category(
        data_x: pd.DataFrame,
        data_y: pd.DataFrame,
        categorical_columns: list,
        output_dir: Path
    ) -> None:
    """
    Plots Kaplan-Meier survival curves for each category in the specified categorical columns.

    Args:
        data_x (pd.DataFrame): Dataset containing the categorical columns to group by.
        data_y (pd.DataFrame): Dataset containing 'time' and 'event' columns for survival analysis.
        categorical_columns (list): List of categorical column names to generate survival curves for.
        output_dir (Path): Directory to save the Kaplan-Meier survival curve plots.
    """
    output_dir.mkdir(parents=True, exist_ok=True)

    for cat_col in categorical_columns:
        plt.figure(figsize=(10, 6))
        plt.title(f"Kaplan-Meier Survival Curve by {cat_col}")

        unique_categories = data_x[cat_col].unique()

        # Plot survival curves for each category
        for category in unique_categories:
            mask_category = data_x[cat_col] == category
            try: # To catch when there are not enough samples for category
                time_category, survival_prob_category, conf_int = kaplan_meier_estimator(
                    data_y["event"][mask_category].astype(bool),
                    data_y["time"][mask_category],
                    conf_type="log-log",
                )

                plt.step(
                    time_category,
                    survival_prob_category,
                    where="post",
                    label=f"{cat_col} = {category}"
                )
                plt.fill_between(
                    time_category,
                    conf_int[0],
                    conf_int[1],
                    alpha=0.25,
                    step="post"
                )
            except Exception as _:
                pass
        
        results_multivariate = multivariate_logrank_test(
            data_y['time'], 
            data_x[cat_col], 
            data_y['event']
        )
        multivariate_p_value = results_multivariate.p_value

        plt.text(0.6, 0.1, f"Multivariate log-rank p-value: {multivariate_p_value:.4e}",
                 fontsize=10, transform=plt.gca().transAxes, bbox=dict(facecolor='white', alpha=0.8))
        plt.ylim(0, 1)
        plt.ylabel(r"Estimated Probability of Survival $\hat{S}(t)$")
        plt.xlabel("Time $t$")
        plt.legend(loc="best")
        plt.grid(alpha=0.3)
        plt.savefig(output_dir / f'kaplan_meier_{cat_col}.png')
        plt.close()

# EXPLAINER

class ModelWrapper:
    def __init__(self, predictor: TabularPredictor, feature_names: list, target_variable: str=None):
        self.ag_model = predictor
        self.feature_names = feature_names
        self.target_variable = target_variable
        if target_variable is None and predictor.problem_type != 'regression':
            print("Since target_class not specified, SHAP will explain predictions for each class")

    def predict_proba(self, X):
        if isinstance(X, pd.Series):
            X = X.values.reshape(1,-1)
        if not isinstance(X, pd.DataFrame):
            X = pd.DataFrame(X, columns=self.feature_names)

        if self.ag_model.can_predict_proba:
            preds = self.ag_model.predict_proba(X)
        else:
            preds = self.ag_model.predict(X)
        return preds

@config_plot()
def plot_feature_importance(df: pd.DataFrame, output_dir: Path, model_name: str=''):
    """
    Plots feature importance with standard deviation and p-value significance.

    Args:
        df (pd.DataFrame): DataFrame containing the feature importance data. 
            Look at example for required format.
        output_dir (Path): Directory to save the feature importance plot.
        model_name (str): Optional name of the model, included in the plot title.

    Example:
        ```python
        import pandas as pd
        from pathlib import Path

        df = pd.DataFrame({
            'importance': [0.25, 0.18, 0.12, 0.10],
            'stddev': [0.03, 0.02, 0.01, 0.015],
            'p_value': [0.03, 0.07, 0.01, 0.2]
        }, index=['Feature A', 'Feature B', 'Feature C', 'Feature D'])

        plot_feature_importance(df, Path('./output'))
        ```
    """
    fig, ax = plt.subplots(figsize=(20, 12), dpi=72)

    bars = ax.bar(df.index, df['importance'], yerr=df['stddev'], capsize=5, color='skyblue', edgecolor='black')

    if 'p_value' in df.columns:
        for bar, p_value in zip(bars, df['p_value']):
            height = bar.get_height()
            significance = '*' if p_value < 0.05 else ''
            ax.text(bar.get_x() + bar.get_width() / 2.0, height + 0.002, significance,
                    ha='center', va='bottom', fontsize=10, color='red')

    ax.set_xlabel('Feature', fontsize=14)
    ax.set_ylabel('Importance', fontsize=14)
    ax.set_title(f'Feature Importance with Standard Deviation and p-value Significance ({model_name})', fontsize=16)
    ax.axhline(0, color='grey', linewidth=0.8)

    ax.set_xticks(np.arange(len(df.index.values)))
    ax.set_xticklabels(df.index.values, rotation=60, ha='right', fontsize=10)

    ax.grid(axis='y', linestyle='--', alpha=0.7)

    significance_patch = plt.Line2D([0], [0], color='red', marker='*', linestyle='None', markersize=10, label='p < 0.05')
    ax.legend(handles=[significance_patch], loc='upper right', fontsize=12)

    plt.tight_layout()
    fig.savefig(output_dir / 'feature_importance.png')
    plt.close()

@config_plot()
def plot_shap_values(
        predictor: TabularPredictor,
        X_train: pd.DataFrame,
        X_test: pd.DataFrame,
        output_dir: Path,
        max_display: int = 10,
    ) -> None:
    """
    Generates and saves SHAP value visualizations, including a heatmap and a bar plot, for a autogluon tabular model.

    Args:
        predictor (TabularPredictor): The trained tabular predictor model for which SHAP values are calculated.
        X_train (pd.DataFrame): Training dataset used to create the SHAP background data.
        X_test (pd.DataFrame): Test dataset used to evaluate and compute SHAP values.
        output_dir (Path): Directory to save the SHAP value visualizations.
        max_display (int): Maximum number of features to display in the visualizations. Defaults to 10.
    """
    predictor = ModelWrapper(predictor, X_train.columns)
    background_data = shap.sample(X_train, 100)
    shap_exp = shap.KernelExplainer(predictor.predict_proba, background_data)

    # sample 100 samples from test set to evaluate with shap values
    test_data = shap.sample(X_test, 100)

    # Compute SHAP values for the test set
    shap_values = shap_exp(test_data)

    sns.set_theme(style="darkgrid", font="Arial")
    fig, ax = plt.subplots(figsize=(20, 12), dpi=72)
    shap.plots.heatmap(shap_values[...,1], max_display=max_display, show=False, ax=ax)
    fig.savefig(output_dir / 'shap_heatmap.png')
    plt.close()

    fig, ax = plt.subplots(figsize=(20, 12), dpi=72)
    shap.plots.bar(shap_values[...,1], max_display=max_display, show=False, ax=ax)
    fig.savefig(output_dir / 'shap_barplot.png')
    plt.close()

def plot_violin_of_bootstrapped_metrics(
        trainer,
        X_test: pd.Series,
        y_test: pd.Series,
        X_val: pd.Series,
        y_val: pd.Series,
        X_train: pd.Series,
        y_train: pd.Series,
        output_dir: Path
    ) -> None:
    """
    Generates violin plots for bootstrapped model performance metrics across train, validation, and test datasets.

    Args:
        trainer (TrainerSupervised): The trained model predictor for evaluating performance metrics.
        X_test (pd.Series): Test features dataset.
        y_test (pd.Series): Test target values.
        X_val (pd.Series): Validation features dataset.
        y_val (pd.Series): Validation target values.
        X_train (pd.Series): Training features dataset.
        y_train (pd.Series): Training target values.
        output_dir (Path): Directory to save the generated violin plots.
    """
    # Define metrics based on the problem type
    if trainer.task == 'regression':
        metrics = [('R Squared', r2_score), ('Root Mean Squared Error', root_mean_squared_error)]
    elif trainer.task == 'binary':
        metrics = [('AUROC', roc_auc_score), ('AUPRC', auprc)]
    elif trainer.task == 'survival':
        metrics = [('Concordance Index', ci_wrapper)]

    # Prepare lists for DataFrame
    results = []

    # Loop through models and metrics to compute bootstrapped values
    for model_name in trainer.model_names():
        y_pred_test = trainer.infer(X_test, model=model_name)
        y_pred_val = trainer.infer(X_val, model=model_name)
        y_pred_train = trainer.infer(X_train, model=model_name)

        for metric_name, metric_func in metrics:
            test_values = bootstrap_metric(y_test.to_numpy(), y_pred_test, metric_func)
            results.extend([(model_name, metric_name, 'Test', value) for value in test_values])

            val_values = bootstrap_metric(y_val.to_numpy(), y_pred_val, metric_func)
            results.extend([(model_name, metric_name, 'Validation', value) for value in val_values])

            train_values = bootstrap_metric(y_train.to_numpy(), y_pred_train, metric_func)
            results.extend([(model_name, metric_name, 'Train', value) for value in train_values])

    # Create a results DataFrame
    result_df = pd.DataFrame(results, columns=['model', 'metric', 'data_split', 'value'])

     # Sort models by median metric value within each combination of metric and data_split
    model_order_per_split = {}
    for split in ['Test', 'Validation', 'Train']:
        split_order = (
            result_df[result_df['data_split'] == split]
            .groupby(['metric', 'model'])['value']
            .median()
            .reset_index()
            .sort_values(by=['metric', 'value'], ascending=[True, False])
            .groupby('metric')['model']
            .apply(list)
            .to_dict()
        )
        model_order_per_split[split] = split_order

    # Function to create violin plots for a specific data split
    def create_violin_plot(data_split, save_path):
        sns.set_theme(style="darkgrid", font="Arial")
        subset = result_df[result_df['data_split'] == data_split]
        g = sns.FacetGrid(
            subset,
            col="metric",
            margin_titles=True,
            height=4,
            aspect=1.5,
            sharex=False,
        )

        # Create violin plots with sorted models
        def violin_plot(data, **kwargs):
            metric = data.iloc[0]['metric']
            order = model_order_per_split[data_split].get(metric, None)
            sns.violinplot(data=data, x="value", y="model", linewidth=1, order=order, **kwargs)

        g.map_dataframe(violin_plot)

        # Adjust the titles and axis labels
        g.set_titles(col_template="{col_name}")
        g.set_axis_labels("", "Model")

        # Add overall title and adjust layout
        g.figure.suptitle(f"Model Performance of {data_split} Data (Bootstrapped)", fontsize=16)
        g.tight_layout(w_pad=0.5, h_pad=1)

        # Save the plot
        g.savefig(save_path, dpi=500)
        plt.close()

    # Generate and save plots for each data split
    create_violin_plot('Test', output_dir / 'test_metrics_bootstrap.png')
    create_violin_plot('Validation', output_dir / 'validation_metrics_bootstrap.png')
    create_violin_plot('Train', output_dir / 'train_metrics_bootstrap.png')

@config_plot()
def plot_regression_diagnostics(
        y_true: np.ndarray, 
        y_pred: np.ndarray, 
        output_dir: Path
    ) -> None:
    """
    Generates diagnostic plots for evaluating a regression model.

    Plots:
        - True vs. Predicted values plot.
        - Residuals plot.
        - Histogram of residuals.

    Args:
        y_true (np.ndarray): Array of true target values.
        y_pred (np.ndarray): Array of predicted values from the regression model.
        output_dir (Path): Directory to save the diagnostic plots.
    """
    residuals = y_true - y_pred
    
    # Regression Line
    plt.figure(figsize=(10, 6))
    sns.set_theme(style="darkgrid", font="Arial")
    sns.scatterplot(x=y_true, y=y_pred, alpha=0.5)
    sns.lineplot(x=y_true, y=y_true, color='red')  # Perfect prediction line
    plt.xlabel('True Values')
    plt.ylabel('Predictions')
    plt.title('True vs Predicted Values')
    plt.savefig(output_dir / 'true_vs_predicted.png')
    plt.close()

    # Residuals
    plt.figure(figsize=(10, 6))
    sns.set_theme(style="darkgrid", font="Arial")
    sns.scatterplot(x=y_pred, y=residuals, alpha=0.5)
    plt.axhline(0, color='red', linestyle='--')
    plt.xlabel('Fitted Values')
    plt.ylabel('Residuals')
    plt.title('Residual Plot')
    plt.savefig(output_dir / 'residual_plot.png')
    plt.close()

    # Residual Histogram
    plt.figure(figsize=(10, 6))
    sns.set_theme(style="darkgrid", font="Arial")
    sns.histplot(residuals, kde=True, bins=30)
    plt.xlabel('Residuals')
    plt.title('Histogram of Residuals')
    plt.savefig(output_dir / 'residual_hist.png')
    plt.close()

@config_plot()
def plot_classification_diagnostics(
        y_test: pd.DataFrame,
        y_test_pred: pd.DataFrame,
        y_val: pd.DataFrame,
        y_val_pred: pd.DataFrame,
        y_train: pd.DataFrame,
        y_train_pred: pd.DataFrame,
        output_dir: Path
    ) -> None:
    """
    Generates diagnostic plots for evaluating a classification model.

    Plots:
        - Epic model evaluation plots (ROC Curve, Precision-Recall Curve, Calibration Curve, Sensitivity/Flag Curve).
        - Confusion Matrix.

    Args:
        y_test (pd.DataFrame): True labels for the test dataset.
        y_test_pred (pd.DataFrame): Predicted probabilities for the test dataset.
        y_val (pd.DataFrame): True labels for the validation dataset.
        y_val_pred (pd.DataFrame): Predicted probabilities for the validation dataset.
        y_train (pd.DataFrame): True labels for the training dataset.
        y_train_pred (pd.DataFrame): Predicted probabilities for the training dataset.
        output_dir (Path): Directory to save the diagnostic plots.
    """
    plot_epic_copy(
        y_test.to_numpy(),
        y_test_pred.to_numpy(),
        y_val.to_numpy(),
        y_val_pred.to_numpy(),
        y_train.to_numpy(),
        y_train_pred.to_numpy() ,
        output_dir
    )

    conf_matrix = confusion_matrix(y_test, y_test_pred.apply(lambda x: 1 if x >= 0.5 else 0))

    plt.figure(figsize=(8, 6))
    sns.heatmap(conf_matrix, annot=True, fmt='d', cmap='Blues')
    plt.xlabel('Predicted')
    plt.ylabel('Actual')
    plt.title('Confusion Matrix')
    plt.savefig(output_dir / 'confusion_matrix.png')
    plt.close()




from fpdf import FPDF
from fpdf.enums import YPos
import pandas as pd

# UTILS
def _add_multiplots(pdf: FPDF, multiplots: list, categorical_columns: list, continuous_columns: list) -> FPDF:
    n_continuous = len(continuous_columns) + 2
    if n_continuous < 11:
        n_rows = 19 // n_continuous
    else:
        n_rows = 1

    # start first page of multiplots
    current_y = pdf.t_margin
    for n, (plot, cat) in enumerate(zip(multiplots, categorical_columns)):
        if n % n_rows == 0:
            pdf.add_page()
            current_y = pdf.get_y()
        
        pdf.set_font('inter', 'B', 24)
        pdf.set_y(current_y)
        pdf.cell(text=f"{cat.title()} Multiplots", new_y=YPos.NEXT)

        current_y = pdf.get_y() + 2

        img_width = pdf.epw
        img = pdf.image(
            plot, 
            x=pdf.l_margin, 
            y=current_y, 
            w=img_width, 
            keep_aspect_ratio=True
        )
        current_y += img.rendered_height + 6
        pdf.set_y(current_y)    

    return pdf

def _add_tableone(pdf: FPDF, data: pd.DataFrame) -> FPDF:
    # Keep empty header entries
    data.columns = [f"col_{n}" if 'Unnamed:' in header else header for n, header in enumerate(data.columns)]
    
    # save 
    n = data.iloc[0]['Overall']
    data = data.iloc[1:]
    
    continuous_columns = [col.replace(', mean (SD)', '') for col in data['col_0'].unique() if 'mean (SD)' in col]
    categorical_columns = [col.replace(', n (%)', '') for col in data['col_0'].unique() if 'n (%)' in col]

    # new page for continuous variables + title 
    pdf.add_page()
    pdf.set_font('inter', 'B', 24)
    pdf.cell(h=pdf.t_margin, text='Table 1: Continuous Variables', new_y=YPos.NEXT)

    # table starts here
    pdf.set_font('inter', '', 10)
    with pdf.table(col_widths=(1.5, 1, 1, 1), text_align=['LEFT', 'RIGHT', 'RIGHT', 'RIGHT']) as table:
        # header row
        row = table.row()
        row.cell('Feature Name')
        row.cell(f'Missing (n={n})')
        row.cell('Mean')
        row.cell('SD')

        # iterate through each continuous variable and render its data on a table
        for col in continuous_columns:
            data_row = data[data["col_0"].str.startswith(f"{col},")].iloc[0]

            # separate mean/sd columns
            data_row['Mean'] = data_row['Overall'].split(" (")[0]
            data_row['SD'] = data_row['Overall'].split(" (")[1].split(")")[0]

            # one row of continuous variable
            row = table.row()
            row.cell(col)
            row.cell(data_row['Missing'])
            row.cell(data_row['Mean'])
            row.cell(data_row['SD'])
    
    # start next table 10mm after
    pdf.set_y(pdf.get_y() + 10)
    pdf.set_font('inter', 'B', 24)
    pdf.cell(h=pdf.t_margin, text='Table 2: Categorical Variables', new_y=YPos.NEXT)

    # table starts here
    pdf.set_font('inter', '', 10)
    with pdf.table(col_widths=(1.5, 1, 1, 1), text_align=['LEFT', 'LEFT', 'RIGHT', 'RIGHT']) as table:
        # header row
        row = table.row()
        row.cell('Feature Name')
        row.cell('Value')
        row.cell(f'Count (n={n})')
        row.cell('%')

        # iterate through each categorical variable and render its data on a table
        for col in categorical_columns:
            rows = data[data["col_0"].str.startswith(f"{col},")].reset_index(drop=True)
            
            # separate mean/sd columns
            rows['n'] = rows['Overall'].apply(lambda x: x.split(" (")[0])
            rows['%'] = rows['Overall'].apply(lambda x: x.split(" (")[1].split(")")[0])
    
            # iterate through each value of the categorical variable
            for n, data_row in rows.iterrows():
                row = table.row()
                if n == 0:
                    row.cell(col, rowspan=len(rows))
                row.cell(data_row['col_1'])
                row.cell(data_row['n'])
                row.cell(data_row['%'])

    return pdf

def _add_outlier_analysis(pdf: FPDF, outlier_analysis: str) -> FPDF:
    from ast import literal_eval
    pdf.set_font('inter', '', 10)

    outliers = {}
    for line in outlier_analysis.splitlines():
        var = line.split("found in")[1].split(": [")[0].strip()
        if "No Outliers" in line:
            outliers[var] = "✓ No outliers found"
        else:
            outliers[var] = literal_eval(line.split(f"{var}: ")[1])
    
    # table starts here
    pdf.set_font('inter', '', 10)
    with pdf.table(col_widths=(1, 2), text_align=['LEFT', 'LEFT']) as table:
        # header row
        row = table.row()
        row.cell('Feature Name')
        row.cell('Outlier')

        # iterate through each categorical variable and render its data on a table
        for var in outliers:
            var_outs = outliers[var]
            if isinstance(var_outs, str):
                row = table.row()
                row.cell(var)
                row.cell(var_outs)
            else:
                for n, val in enumerate(var_outs): 
                    row = table.row()
                    if n == 0:
                        row.cell(var, rowspan=len(var_outs))
                    row.cell(val)

    return pdf



from pathlib import Path

import pandas as pd
import numpy as np

from tabulate import tabulate

from sklearn.metrics import log_loss

from .subgroup_analysis import subgroup_analysis_fairlearn, subgroup_analysis_OLS
from ..utils import plot_violin

class ModelAuditor():
    """
    A class for explaining and analyzing bias in a predictive model's outcomes based on sensitive features.
    """
    def __init__(
            self, 
            y_true: pd.Series, 
            y_pred: np.ndarray, 
            sensitive_features: pd.DataFrame, 
            task: str,
            output_dir: Path,
            **kwargs: dict
        ) -> None:
        self.y_true = y_true
        self.y_pred = y_pred
        self.sensitive_features = sensitive_features
        self.task = task
        self.output_dir = output_dir
        self.mapper = {"mean_prediction": "Demographic Parity",
                       "false_positive_rate": "(FPR) Equalized Odds",
                       "true_positive_rate": "(TPR) Equalized Odds or Equal Opportunity"}
        self.kwargs = kwargs
    
    def run(
            self, 
            relative: bool = False, 
            fairness_threshold: float = 1.2
        ) -> None:
        """
        Runs the bias explainer analysis on the provided data. It first evaluates the potential bias in the model's predictions
        using the OLS regression F-statistic p-value. If the p-value is below the threshold of 0.05, indicating 
        potential bias in the sensitive feature, the method proceeds to generate visualizations and calculate fairness metrics.

        Args:
            relative (bool): 
                If True, the metrics will be presented relative to the most frequent value of each sensitive feature.
            fairness_threshold (float): 
                A threshold for determining fairness based on relative metrics. If the relative metric exceeds this threshold, 
                a warning flag will be applied.
        """
        if self.task == 'binary':
            y_true_array = self.y_true.to_numpy()
            bias_metric = np.array([
                log_loss([y_true_array[idx]], [self.y_pred[idx]], labels=np.unique(y_true_array))
                for idx in range(len(y_true_array))
            ])
            self.y_pred = (self.y_pred >= .5).astype(int)
        elif self.task == 'regression':
            bias_metric = np.sqrt((self.y_true.to_numpy() - self.y_pred) ** 2)

        self.results = []
        for sensitive_feature in self.sensitive_features.columns:
            _, f_pvalue = subgroup_analysis_OLS(sensitive_feature, bias_metric, self.output_dir)
            if f_pvalue < 0.05:
                plot_violin(sensitive_feature, bias_metric, self.output_dir, show_figure=True)
                result = subgroup_analysis_fairlearn(sensitive_feature, fairness_threshold, relative=relative)

                print(f"\n=== Subgroup Analysis for '{sensitive_feature.title()}' using FairLearn ===\n")
                table_output = tabulate(result.iloc[:, :4], headers='keys', tablefmt='grid')
                print('\n'.join(['    ' + line for line in table_output.split('\n')]), '\n')

                result.to_csv(self.output_dir / f'{sensitive_feature}_fm_metrics.csv')



# cancer site / cancer morphology
cancer_code_map = {
    'C00': 'Lip',
    'C01': 'Base of tongue',
    'C02': 'Other and unspecified parts of tongue',
    'C03': 'Gum',
    'C04': 'Floor of mouth',
    'C05': 'Palate',
    'C06': 'Other and unspecified parts of mouth',
    'C07': 'Parotid gland',
    'C08': 'Other and unspecified major salivary glands',
    'C09': 'Tonsil',
    'C10': 'Oropharynx',
    'C11': 'Nasopharynx',
    'C12': 'Pyriform sinus',
    'C13': 'Hypopharynx',
    'C14': 'Other and ill-defined sites in lip, oral cavity, and pharynx',
    'C15': 'Esophagus',
    'C16': 'Stomach',
    'C17': 'Small intestine',
    'C18': 'Colon',
    'C19': 'Rectosigmoid junction',
    'C20': 'Rectum',
    'C21': 'Anus and anal canal',
    'C22': 'Liver and intrahepatic bile ducts',
    'C23': 'Gallbladder',
    'C24': 'Other and unspecified parts of biliary tract',
    'C25': 'Pancreas',
    'C26': 'Other and ill-defined digestive organs',
    'C30': 'Nasal cavity and middle ear',
    'C31': 'Accessory sinuses',
    'C32': 'Larynx',
    'C33': 'Trachea',
    'C34': 'Bronchus and lung',
    'C37': 'Thymus',
    'C38': 'Heart, mediastinum, and pleura',
    'C40': 'Bones, joints, and articular cartilage of limbs',
    'C41': 'Bones, joints, and articular cartilage of other and unspecified sites',
    'C44': 'Skin',
    'C47': 'Peripheral nerves and autonomic nervous system',
    'C48': 'Retroperitoneum and peritoneum',
    'C49': 'Connective, subcataneous, and other soft tissues',
    'C50': 'Breast', 
    'C51': 'Vulva',
    'C52': 'Vagina',
    'C53': 'Cervix uteri',
    'C54': 'Corpus uteri',
    'C55': 'Uterus, NOS',
    'C56': 'Ovary',
    'C57': 'Other and unspecified female genital organs',
    'C58': 'Placenta',
    'C60': 'Penis',
    'C61': 'Prostate gland',
    'C62': 'Testis',
    'C63': 'Other and unspecified male genital organs',
    'C64': 'Kidney',
    'C65': 'Renal pelvis',
    'C66': 'Ureter',
    'C67': 'Bladder',
    'C68': 'Other and unspecified urinary organs',
    'C69': 'Eye and adnexa',
    'C71': 'Brain',
    'C72': 'Spinal cord, cranial nerves, and other parts of central nervous system',
    'C73': 'Thyroid and other endocrine glands',
    'C74': 'Adrenal Gland', 
    'C75': 'Other endocrine glands and related structures', 
    'C76': 'Other and ill-defined sites',
    'C80': 'Unknown primary site',
    '800': 'Neoplasms, NOS',
    '801': 'Epithelial neoplasms, NOS',
    '802': 'Epithelial neoplasms, NOS',
    '803': 'Epithelial neoplasms, NOS',
    '804': 'Epithelial neoplasms, NOS',
    '805': 'Squamous cell neoplasms',
    '807': 'Squamous cell neoplasms',
    '808': 'Squamous cell neoplasms',
    '809': 'Basal cell neoplasms',
    '812': 'Transitional cell papillomas and carcinomas',
    '813': 'Transitional cell papillomas and carcinomas',
    '814': 'Adenomas and adenocarcinomas',
    '816': 'Adenomas and adenocarcinomas',
    '817': 'Adenomas and adenocarcinomas',
    '818': 'Adenomas and adenocarcinomas',
    '820': 'Adenomas and adenocarcinomas',
    '821': 'Adenomas and adenocarcinomas',
    '823': 'Adenomas and adenocarcinomas',
    '824': 'Adenomas and adenocarcinomas',
    '825': 'Adenomas and adenocarcinomas',
    '826': 'Adenomas and adenocarcinomas',
    '829': 'Adenomas and adenocarcinomas',
    '831': 'Adenomas and adenocarcinomas',
    '832': 'Adenomas and adenocarcinomas',
    '833': 'Adenomas and adenocarcinomas',
    '834': 'Adenomas and adenocarcinomas',
    '837': 'Adenomas and adenocarcinomas',
    '838': 'Adenomas and adenocarcinomas',
    '840': 'Adnexal and skin appendage neoplasms',
    '843': 'Mucoepidermoid neoplasms',
    '844': 'Cystic, mucinous, and serous neoplasms',
    '845': 'Cystic, mucinous, and serous neoplasms',
    '846': 'Cystic, mucinous, and serous neoplasms',
    '847': 'Cystic, mucinous, and serous neoplasms',
    '848': 'Cystic, mucinous, and serous neoplasms',
    '849': 'Cystic, mucinous, and serous neoplasms',
    '850': 'Ductal and lobular neoplasms',
    '851': 'Ductal and lobular neoplasms',
    '852': 'Ductal and lobular neoplasms',
    '853': 'Ductal and lobular neoplasms',
    '854': 'Ductal and lobular neoplasms',
    '855': 'Acinar cell neoplasms',
    '856': 'Complex epithelial neoplasms',
    '857': 'Complex epithelial neoplasms',
    '858': 'Thymic epithelial neoplasms',
    '862': 'Specialized gonadal neoplasms',
    '872': 'Paragangliomas and glomus tumors',
    '873': 'Nevi and melanomas',
    '874': 'Nevi and melanomas',
    '877': 'Nevi and melanomas',
    '880': 'Soft tissue tumors and sarcomas, NOS',
    '881': 'Fibromatous neoplasms',
    '883': 'Fibromatous neoplasms',
    '885': 'Lipomatous neoplasms',
    '889': 'Myomatous neoplasms',
    '890': 'Myomatous neoplasms',
    '891': 'Myomatous neoplasms',
    '892': 'Myomatous neoplasms',
    '892': 'Myomatous neoplasms',
    '893': 'Complex mixed and stromal neoplasms',
    '894': 'Complex mixed and stromal neoplasms',
    '895': 'Complex mixed and stromal neoplasms',
    '898': 'Complex mixed and stromal neoplasms',
    '902': 'Fibroepithelial neoplasms',
    '904': 'Synovial-like neoplasms',
    '905': 'Mesothelial neoplasms',
    '906': 'Germ cell neoplasms',
    '907': 'Germ cell neoplasms',
    '908': 'Germ cell neoplasms',
    '910': 'Trophoblastic neoplasms',
    '911': 'Mesonephromas',
    '912': 'Blood vessel tumors',
    '914': 'Blood vessel tumors',
    '915': 'Blood vessel tumors',
    '918': 'Osseous and chondromatous neoplasms',
    '922': 'Osseous and chondromatous neoplasms',
    '924': 'Osseous and chondromatous neoplasms',
    '926': 'Miscellaneous bone tumors',
    '936': 'Miscellaneous tumors',
    '938': 'Gliomas',
    '940': 'Gliomas',
    '942': 'Gliomas',
    '944': 'Gliomas',
    '945': 'Gliomas',
    '947': 'Gliomas',
    '954': 'Nerve sheath tumors',
    '956': 'Nerve sheath tumors',
}

# lab observations
obs_map = {
    'Hematology': {
        'Basophils': 'basophil',
        'Basos': 'basophil',
        'Eosinophils': 'eosinophil',
        'Eosin': 'eosinophil',
        'Eos': 'eosinophil',
        'Lymphocytes': 'lymphocyte',
        'Lymphs': 'lymphocyte',
        'Lymph': 'lymphocyte',
        'Monocytes': 'monocyte', 
        'Monos': 'monocyte',
        'Mono': 'monocyte', 
        'Neutrophils': 'neutrophil',
        
        # NOTE: Hb also exists in Biochemistry, but the procedure is different (arterial blood gas test instead of CBC)
        'Hb': 'hemoglobin',
        'Hct': 'hematocrit',
        'Plt': 'platelet',
        
        'MCH': 'mean_corpuscular_hemoglobin',
        'MCHC': 'mean_corpuscular_hemoglobin_concentration',
        'MCV': 'mean_corpuscular_volume',
        'MPV': 'mean_platelet_volume',
        'Mean Platelet Volume': 'mean_platelet_volume',
        'RBC': 'red_blood_cell',
        'RDW': 'red_cell_distribution_width',
        'WBC': 'white_blood_cell',

        'INR': 'prothrombin_time_international_normalized_ratio',
        'aPTT': 'activated_partial_thromboplastin_time',
        'APTT': 'activated_partial_thromboplastin_time',
    },
    'Biochemistry': {
        'Albumin': 'albumin',
        'Bicarbonate': 'bicarbonate',
        'Carcinoembryonic Antigen': 'carcinoembryonic_antigen',
        'Chloride': 'chloride',
        'Creatinine': 'creatinine',
        'Glucose': 'glucose',
        'Glucose, Random': 'glucose',
        'Magnesium': 'magnesium',
        'Phosphate': 'phosphate',
        'Potassium': 'potassium',
        'Sodium': 'sodium',
        'Total Bilirubin': 'total_bilirubin',
        'Tot Bilirubin': 'total_bilirubin',
        'Bilirubin,Total': 'total_bilirubin',
        'RDW-CV': 'red_cell_distribution_width',

        'AST': 'aspartate_aminotransferase',
        'ALT': 'alanine_aminotransferase',
        'ALP': 'alkaline_phosphatase',
        'Alkaline Phosphatase': 'alkaline_phosphatase',
        'CA 19-9': 'carbohydrate_antigen_19-9',
        'LDH': 'lactate_dehydrogenase',
        'Lactate Dehydrogenase': 'lactate_dehydrogenase',
    }
}

symp_cols = [
    'esas_pain',
    'esas_tiredness',
    'esas_nausea',
    'esas_depression',
    'esas_anxiety',
    'esas_drowsiness',
    'esas_appetite',
    'esas_well_being',
    'esas_shortness_of_breath',
    'patient_ecog',
]
lab_cols = [
    'alanine_aminotransferase',
    'albumin',
    'alkaline_phosphatase',
    'aspartate_aminotransferase',
    'basophil',
    'bicarbonate',
    'chloride',
    'creatinine',
    'eosinophil',
    'glucose',
    'hematocrit',
    'hemoglobin',
    'lactate_dehydrogenase',
    'lymphocyte',
    'magnesium',
    'mean_corpuscular_hemoglobin',
    'mean_corpuscular_hemoglobin_concentration',
    'mean_corpuscular_volume',
    'mean_platelet_volume',
    'monocyte',
    'neutrophil',
    'phosphate',
    'platelet',
    'potassium',
    'red_blood_cell',
    'red_cell_distribution_width',
    'sodium',
    'total_bilirubin',
    'white_blood_cell',
]
symp_change_cols = [f'{col}_change' for col in symp_cols]
lab_change_cols = [f'{col}_change' for col in lab_cols]
drug_cols = [
    '%_ideal_dose_given_CISPLATIN',
    '%_ideal_dose_given_FLUOROURACIL', 
    '%_ideal_dose_given_ETOPOSIDE',
    '%_ideal_dose_given_CARBOPLATIN', 
    '%_ideal_dose_given_OXALIPLATIN',
    '%_ideal_dose_given_PEMBROLIZUMAB', 
    '%_ideal_dose_given_NIVOLUMAB',
    '%_ideal_dose_given_DOCETAXEL', 
    '%_ideal_dose_given_CYCLOPHOSPHAMIDE',
    '%_ideal_dose_given_DURVALUMAB', 
    '%_ideal_dose_given_CETUXIMAB',
    '%_ideal_dose_given_RAMUCIRUMAB', 
    '%_ideal_dose_given_RALTITREXED',
    '%_ideal_dose_given_IPILIMUMAB', 
    '%_ideal_dose_given_MITOMYCIN',
    '%_ideal_dose_given_PANITUMUMAB', 
    '%_ideal_dose_given_CAPECITABINE',
    '%_ideal_dose_given_ERLOTINIB',
]